#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Unit tests for schooltool.calendar.icalendar
"""

import unittest
import doctest
import difflib
import time
import os
from pprint import pformat, pprint
from textwrap import dedent
from datetime import date, timedelta, datetime
from StringIO import StringIO


def diff(old, new, oldlabel="expected output", newlabel="actual output"):
    """Display a unified diff between old text and new text."""
    old = old.splitlines()
    new = new.splitlines()

    diff = ['--- %s' % oldlabel, '+++ %s' % newlabel]

    def dump(tag, x, lo, hi):
        for i in xrange(lo, hi):
            diff.append(tag + x[i])

    differ = difflib.SequenceMatcher(a=old, b=new)
    for tag, alo, ahi, blo, bhi in differ.get_opcodes():
        if tag == 'replace':
            dump('-', old, alo, ahi)
            dump('+', new, blo, bhi)
        elif tag == 'delete':
            dump('-', old, alo, ahi)
        elif tag == 'insert':
            dump('+', new, blo, bhi)
        elif tag == 'equal':
            dump(' ', old, alo, ahi)
        else:
            raise AssertionError('unknown tag %r' % tag)
    return "\n".join(diff)


class TimezoneTestMixin:
    """A mixin for tests that fiddle with timezones."""

    def setUp(self):
        self.have_tzset = hasattr(time, 'tzset')
        self.touched_tz = False
        self.old_tz = os.getenv('TZ')

    def tearDown(self):
        if self.touched_tz:
            self.setTZ(self.old_tz)

    def setTZ(self, tz):
        self.touched_tz = True
        if tz is None:
            os.unsetenv('TZ')
        else:
            os.putenv('TZ', tz)
        time.tzset()


class TestParseDateTime(TimezoneTestMixin, unittest.TestCase):

    def test_timezones(self):
        # The simple tests are in the doctest of parse_date_time.
        from schooltool.calendar.icalendar import parse_date_time
        from pytz import utc

        if not self.have_tzset:
            return # Do not run this test on Windows

        self.setTZ('UTC')
        dt = parse_date_time('20041029T125031Z')
        self.assertEquals(dt, datetime(2004, 10, 29, 12, 50, 31, tzinfo=utc))

        self.setTZ('EET-2EEST')
        dt = parse_date_time('20041029T095031Z') # daylight savings
        self.assertEquals(dt, datetime(2004, 10, 29, 9, 50, 31, tzinfo=utc))
        dt = parse_date_time('20041129T095031Z') # no daylight savings
        self.assertEquals(dt, datetime(2004, 11, 29, 9, 50, 31, tzinfo=utc))


class TestPeriod(unittest.TestCase):

    def test(self):
        from schooltool.calendar.icalendar import Period
        dt1 = datetime(2001, 2, 3, 14, 30, 5)
        dt2 = datetime(2001, 2, 3, 16, 35, 20)
        td = dt2 - dt1
        p1 = Period(dt1, dt2)
        self.assertEquals(p1.start, dt1)
        self.assertEquals(p1.end, dt2)

        p2 = Period(dt1, td)
        self.assertEquals(p2.start, dt1)
        self.assertEquals(p2.end, dt2)

        self.assertEquals(p1, p2)

        p = Period(dt1, timedelta(0))
        self.assertEquals(p.start, dt1)
        self.assertEquals(p.end, dt1)

        self.assertRaises(ValueError, Period, dt2, dt1)
        self.assertRaises(ValueError, Period, dt1, -td)

    def test_overlap(self):
        from schooltool.calendar.icalendar import Period
        p1 = Period(datetime(2004, 1, 1, 12, 0), timedelta(hours=1))
        p2 = Period(datetime(2004, 1, 1, 11, 30), timedelta(hours=1))
        p3 = Period(datetime(2004, 1, 1, 12, 30), timedelta(hours=1))
        p4 = Period(datetime(2004, 1, 1, 11, 0), timedelta(hours=3))

        self.assert_(p1.overlaps(p2))
        self.assert_(p2.overlaps(p1))

        self.assert_(p1.overlaps(p3))
        self.assert_(p3.overlaps(p1))

        self.assert_(not p2.overlaps(p3))
        self.assert_(not p3.overlaps(p2))

        self.assert_(p1.overlaps(p4))
        self.assert_(p4.overlaps(p1))

        self.assert_(p1.overlaps(p1))

    def test_cmp(self):
        from schooltool.calendar.icalendar import Period
        p1 = Period(datetime(2004, 1, 1, 12, 0), timedelta(hours=1))
        p2 = Period(datetime(2004, 1, 1, 12, 0), timedelta(hours=1))
        p3 = Period(datetime(2004, 1, 1, 12, 0), timedelta(hours=4))

        self.assert_(p1 == p2)
        self.assert_(p1 != p3)
        self.assertRaises(NotImplementedError, p1.__cmp__, "NOT A PERIOD")


class TestVEvent(unittest.TestCase):

    def test_add(self):
        from schooltool.calendar.icalendar import VEventParser, ICalParseError
        vevent = VEventParser()
        value, params = 'bar', {'VALUE': 'TEXT'}
        vevent.add('foo', value, params)
        self.assertEquals(vevent._props, {'FOO': [(value, params)]})
        value2 = 'guug'
        vevent.add('fie', value2)
        self.assertEquals(vevent._props, {'FOO': [(value, params)],
                                          'FIE': [(value2, {})]})
        vevent.add('fie', value, params)
        self.assertEquals(vevent._props, {'FOO': [(value, params)],
                                          'FIE': [(value2, {}),
                                                  (value, params)]})

        # adding a singleton property twice
        vevent.add('uid', '1')
        self.assertRaises(ICalParseError, vevent.add, 'uid', '2')

    def test_hasProp(self):
        from schooltool.calendar.icalendar import VEventParser
        vevent = VEventParser()
        vevent.add('foo', 'bar', {})
        self.assert_(vevent.hasProp('foo'))
        self.assert_(vevent.hasProp('Foo'))
        self.assert_(not vevent.hasProp('baz'))

    def test__getType(self):
        from schooltool.calendar.icalendar import VEventParser
        vevent = VEventParser()
        vevent.add('x-explicit', '', {'VALUE': 'INTEGER'})
        vevent.add('dtstart', 'implicit type', {})
        vevent.add('x-default', '', {})
        self.assertEquals(vevent._getType('x-explicit'), 'INTEGER')
        self.assertEquals(vevent._getType('dtstart'), 'DATE-TIME')
        self.assertEquals(vevent._getType('x-default'), 'TEXT')
        self.assertEquals(vevent._getType('X-Explicit'), 'INTEGER')
        self.assertEquals(vevent._getType('DtStart'), 'DATE-TIME')
        self.assertEquals(vevent._getType('X-Default'), 'TEXT')
        self.assertRaises(KeyError, vevent._getType, 'nonexistent')

    def test_getOne(self):
        from schooltool.calendar.icalendar import VEventParser
        vevent = VEventParser()

        vevent.add('foo', 'bar', {})
        self.assertEquals(vevent.getOne('foo'), 'bar')
        self.assertEquals(vevent.getOne('Foo'), 'bar')
        self.assertEquals(vevent.getOne('baz'), None)
        self.assertEquals(vevent.getOne('baz', 'quux'), 'quux')
        self.assertEquals(vevent.getOne('dtstart', 'quux'), 'quux')

        vevent.add('int-foo', '42', {'VALUE': 'INTEGER'})
        vevent.add('int-bad', 'xyzzy', {'VALUE': 'INTEGER'})
        self.assertEquals(vevent.getOne('int-foo'), 42)
        self.assertEquals(vevent.getOne('Int-Foo'), 42)
        self.assertRaises(ValueError, vevent.getOne, 'int-bad')

        vevent.add('date-foo', '20030405', {'VALUE': 'DATE'})
        vevent.add('date-bad1', '20030405T1234', {'VALUE': 'DATE'})
        vevent.add('date-bad2', '2003', {'VALUE': 'DATE'})
        vevent.add('date-bad3', '200301XX', {'VALUE': 'DATE'})
        self.assertEquals(vevent.getOne('date-Foo'), date(2003, 4, 5))
        self.assertRaises(ValueError, vevent.getOne, 'date-bad1')
        self.assertRaises(ValueError, vevent.getOne, 'date-bad2')
        self.assertRaises(ValueError, vevent.getOne, 'date-bad3')

        vevent.add('datetime-foo1', '20030405T060708', {'VALUE': 'DATE-TIME'})
        vevent.add('datetime-foo2', '20030405T060708', {'VALUE': 'DATE-TIME'})
        vevent.add('datetime-bad1', '20030405T010203444444',
                                                        {'VALUE': 'DATE-TIME'})
        vevent.add('datetime-bad2', '2003', {'VALUE': 'DATE-TIME'})
        self.assertEquals(vevent.getOne('datetime-foo1'),
                          datetime(2003, 4, 5, 6, 7, 8))
        self.assertEquals(vevent.getOne('Datetime-Foo2'),
                          datetime(2003, 4, 5, 6, 7, 8))
        self.assertRaises(ValueError, vevent.getOne, 'datetime-bad1')
        self.assertRaises(ValueError, vevent.getOne, 'datetime-bad2')

        vevent.add('dur-foo1', '+P11D', {'VALUE': 'DURATION'})
        vevent.add('dur-foo2', '-P2W', {'VALUE': 'DURATION'})
        vevent.add('dur-foo3', 'P1DT2H3M4S', {'VALUE': 'DURATION'})
        vevent.add('dur-foo4', 'PT2H', {'VALUE': 'DURATION'})
        vevent.add('dur-bad1', 'xyzzy', {'VALUE': 'DURATION'})
        self.assertEquals(vevent.getOne('dur-foo1'), timedelta(days=11))
        self.assertEquals(vevent.getOne('Dur-Foo2'), -timedelta(weeks=2))
        self.assertEquals(vevent.getOne('Dur-Foo3'),
                          timedelta(days=1, hours=2, minutes=3, seconds=4))
        self.assertEquals(vevent.getOne('DUR-FOO4'), timedelta(hours=2))
        self.assertRaises(ValueError, vevent.getOne, 'dur-bad1')

        vevent.add('unknown', 'magic', {'VALUE': 'UNKNOWN-TYPE'})
        self.assertEquals(vevent.getOne('unknown'), 'magic')

    def test_validate_error_cases(self):
        from schooltool.calendar.icalendar import VEventParser, ICalParseError

        vevent = VEventParser()
        self.assertRaises(ICalParseError, vevent.validate)

        vevent = VEventParser()
        vevent.add('dtstart', 'xyzzy', {'VALUE': 'TEXT'})
        self.assertRaises(ICalParseError, vevent.validate)

        vevent = VEventParser()
        vevent.add('dtstart', '20010203', {'VALUE': 'DATE'})
        vevent.add('dtend', '20010203T0000', {'VALUE': 'DATE-TIME'})
        self.assertRaises(ICalParseError, vevent.validate)

        vevent = VEventParser()
        vevent.add('dtstart', '20010203', {'VALUE': 'DATE'})
        vevent.add('dtend', '20010203', {'VALUE': 'DATE'})
        self.assertRaises(ICalParseError, vevent.validate)

        vevent = VEventParser()
        vevent.add('dtstart', '20010203', {'VALUE': 'DATE'})
        self.assertRaises(ICalParseError, vevent.validate)

    def test_validate_all_day_events(self):
        from schooltool.calendar.icalendar import VEventParser, ICalParseError

        parser = VEventParser()
        parser.add('summary', 'An event', {})
        parser.add('uid', 'unique', {})
        parser.add('dtstart', '20010203', {'VALUE': 'DATE'})
        vevent = parser.parse()
        self.assert_(vevent.all_day_event)
        self.assertEquals(vevent.summary, 'An event')
        self.assertEquals(vevent.uid, 'unique')
        self.assertEquals(vevent.dtend, date(2001, 2, 4))

        parser = VEventParser()
        parser.add('summary', 'An\\nevent\\; with backslashes', {})
        parser.add('uid', 'unique2', {})
        parser.add('dtstart', '20010203', {'VALUE': 'DATE'})
        parser.add('dtend', '20010205', {'VALUE': 'DATE'})
        vevent = parser.parse()
        self.assertEquals(vevent.summary, 'An\nevent; with backslashes')
        self.assert_(vevent.all_day_event)
        self.assertEquals(vevent.dtstart, date(2001, 2, 3))
        self.assertEquals(vevent.uid, 'unique2')
        self.assertEquals(vevent.dtend, date(2001, 2, 5))

        parser = VEventParser()
        parser.add('dtstart', '20010203', {'VALUE': 'DATE'})
        parser.add('uid', 'unique3', {})
        parser.add('duration', 'P2D')
        vevent = parser.parse()
        self.assertEquals(vevent.summary, None)
        self.assert_(vevent.all_day_event)
        self.assertEquals(vevent.dtstart, date(2001, 2, 3))
        self.assertEquals(vevent.uid, 'unique3')
        self.assertEquals(vevent.dtend, date(2001, 2, 5))

        parser = VEventParser()
        parser.add('dtstart', '20010203', {'VALUE': 'DATE'})
        parser.add('uid', 'unique4', {})
        parser.add('dtend', '20010201', {'VALUE': 'DATE'})
        self.assertRaises(ICalParseError, parser.parse)

        parser = VEventParser()
        parser.add('dtstart', '20010203', {'VALUE': 'DATE'})
        parser.add('uid', 'unique5', {})
        parser.add('dtend', '20010203', {'VALUE': 'DATE'})
        self.assertRaises(ICalParseError, parser.parse)

    def test_validate_not_all_day_events(self):
        from schooltool.calendar.icalendar import VEventParser, ICalParseError

        parser = VEventParser()
        parser.add('dtstart', '20010203T040506')
        parser.add('uid', 'unique', {})
        vevent = parser.parse()
        self.assert_(not vevent.all_day_event)
        self.assertEquals(vevent.dtstart, datetime(2001, 2, 3, 4, 5, 6))
        self.assertEquals(vevent.dtend, datetime(2001, 2, 3, 4, 5, 6))
        self.assertEquals(vevent.rdates, [])

        parser = VEventParser()
        parser.add('dtstart', '20010203T040000')
        parser.add('uid', 'unique', {})
        parser.add('dtend', '20010204T050102')
        vevent = parser.parse()
        self.assert_(not vevent.all_day_event)
        self.assertEquals(vevent.dtstart, datetime(2001, 2, 3, 4, 0, 0))
        self.assertEquals(vevent.dtend, datetime(2001, 2, 4, 5, 1, 2))

        parser = VEventParser()
        parser.add('dtstart', '20010203T040000')
        parser.add('uid', 'unique', {})
        parser.add('duration', 'P1DT1H1M2S')
        vevent = parser.parse()
        self.assert_(not vevent.all_day_event)
        self.assertEquals(vevent.dtstart, datetime(2001, 2, 3, 4, 0, 0))
        self.assertEquals(vevent.dtend, datetime(2001, 2, 4, 5, 1, 2))

        parser = VEventParser()
        parser.add('dtstart', '20010203T010203')
        parser.add('uid', 'unique', {})
        parser.add('rdate', '20010205T040506')
        parser.add('exdate', '20010206T040506')
        vevent = parser.parse()
        self.assertEquals(vevent.rdates, [datetime(2001, 2, 5, 4, 5, 6)])
        self.assertEquals(vevent.exdates, [datetime(2001, 2, 6, 4, 5, 6)])

        parser = VEventParser()
        parser.add('dtstart', '20010203T010203')
        parser.add('uid', 'unique', {})
        parser.add('exdate', '20010206,20020307', {'VALUE': 'DATE'})
        parser.add('rrule', 'FREQ=DAILY')
        vevent = parser.parse()
        self.assertEquals(vevent.exdates, [date(2001, 2, 6), date(2002, 3, 7)])

        parser = VEventParser()
        parser.add('dtstart', '20010203T010203')
        parser.add('uid', 'unique', {})
        parser.add('dtend', '20010203T010202')
        self.assertRaises(ICalParseError, parser.parse)

    def test_timezones(self):
        from schooltool.calendar.icalendar import VEventParser

        parser = VEventParser()
        parser.add('dtstart', '20010203T040000', params={'TZID': 'Vilnius'})
        parser.add('uid', 'unique', {})
        parser.add('dtend', '20010204T050102', params={'TZID': 'Kaunas'})
        vevent = parser.parse()
        self.assert_(not vevent.all_day_event)
        self.assertEquals(vevent.dtstart, datetime(2001, 2, 3, 4, 0, 0))
        self.assertEquals(vevent.dtstart_tzid, 'Vilnius')
        self.assertEquals(vevent.dtend, datetime(2001, 2, 4, 5, 1, 2))
        self.assertEquals(vevent.dtend_tzid, 'Kaunas')

        parser = VEventParser()
        parser.add('dtstart', '20010203T040000', params={'TZID': 'Vilnius'})
        parser.add('uid', 'unique', {})
        parser.add('duration', 'P1DT1H1M2S')
        vevent = parser.parse()
        self.assert_(not vevent.all_day_event)
        self.assertEquals(vevent.dtstart, datetime(2001, 2, 3, 4, 0, 0))
        self.assertEquals(vevent.dtstart_tzid, 'Vilnius')
        self.assertEquals(vevent.dtend, datetime(2001, 2, 4, 5, 1, 2))
        self.assertEquals(vevent.dtend_tzid, 'Vilnius')

    def test_validate_location(self):
        from schooltool.calendar.icalendar import VEventParser
        parser = VEventParser()
        parser.add('dtstart', '20010203T040506')
        parser.add('uid', 'unique5', {})
        parser.add('location', 'Somewhere')
        vevent = parser.parse()
        self.assertEquals(vevent.location, 'Somewhere')

    def test_validate_description(self):
        from schooltool.calendar.icalendar import VEventParser
        parser = VEventParser()
        parser.add('dtstart', '20010203T040506')
        parser.add('uid', 'unique5', {})
        parser.add('description', 'Some long text')
        vevent = parser.parse()
        self.assertEquals(vevent.description, 'Some long text')

    def test_validate_rrule(self):
        from schooltool.calendar.icalendar import VEventParser
        parser = VEventParser()
        parser.add('dtstart', '20010203T040506')
        parser.add('uid', 'unique5', {})
        parser.add('location', 'Somewhere')
        parser.add('rrule', 'FREQ=DAILY;COUNT=3')
        vevent = parser.parse()

        self.assertEquals(vevent.rrule.interval, 1)
        self.assertEquals(vevent.rrule.count, 3)
        self.assertEquals(vevent.rrule.until, None)
        self.assertEquals(vevent.rrule.exceptions, ())

    def test_validate_rrule_exceptions(self):
        from schooltool.calendar.icalendar import VEventParser
        parser = VEventParser()
        parser.add('dtstart', '20010203T040506')
        parser.add('uid', 'unique5', {})
        parser.add('location', 'Somewhere')
        parser.add('rrule', 'FREQ=MONTHLY;BYDAY=3MO')
        parser.add('exdate', '19960402T010000,19960404T010000')
        parser.add('exdate', '19960406T010000,19960408T010000')
        vevent = parser.parse()

        self.assertEquals(vevent.rrule.interval, 1)
        self.assertEquals(vevent.rrule.count, None)
        self.assertEquals(vevent.rrule.until, None)
        self.assertEquals(vevent.rrule.monthly, 'weekday')
        self.assertEquals(vevent.rrule.exceptions,
                          (date(1996, 4, 2), date(1996, 4, 4),
                           date(1996, 4, 6), date(1996, 4, 8)))
        self.assert_(not isinstance(vevent.rrule.exceptions[0], datetime))

    def test_extractListOfDates(self):
        from schooltool.calendar.icalendar import VEventParser, Period, ICalParseError

        vevent = VEventParser()
        vevent.add('rdate', '20010205T040506')
        vevent.add('rdate', '20010206T040506,20010207T000000')
        vevent.add('rdate', '20010208', {'VALUE': 'DATE'})
        vevent.add('rdate', '20010209T000000/20010210T000000',
                   {'VALUE': 'PERIOD'})
        rdates = vevent._extractListOfDates('RDATE', vevent.rdate_types, False)
        expected = [datetime(2001, 2, 5, 4, 5, 6),
                    datetime(2001, 2, 6, 4, 5, 6),
                    datetime(2001, 2, 7, 0, 0, 0),
                    date(2001, 2, 8),
                    Period(datetime(2001, 2, 9), datetime(2001, 2, 10)),
                   ]
        self.assertEqual(expected, rdates,
                         diff(pformat(expected), pformat(rdates)))

        vevent = VEventParser()
        vevent.add('rdate', '20010205T040506', {'VALUE': 'TEXT'})
        self.assertRaises(ICalParseError, vevent._extractListOfDates, 'RDATE',
                                          vevent.rdate_types, False)

        vevent = VEventParser()
        vevent.add('exdate', '20010205T040506/P1D', {'VALUE': 'PERIOD'})
        self.assertRaises(ICalParseError, vevent._extractListOfDates, 'EXDATE',
                                          vevent.exdate_types, False)

        vevent = VEventParser()
        vevent.add('rdate', '20010208', {'VALUE': 'DATE'})
        rdates = vevent._extractListOfDates('RDATE', vevent.rdate_types, True)
        expected = [date(2001, 2, 8)]
        self.assertEqual(expected, rdates,
                         diff(pformat(expected), pformat(rdates)))

        vevent = VEventParser()
        vevent.add('rdate', '20010205T040506', {'VALUE': 'DATE-TIME'})
        self.assertRaises(ICalParseError, vevent._extractListOfDates, 'RDATE',
                                          vevent.rdate_types, True)

        vevent = VEventParser()
        vevent.add('rdate', '20010209T000000/20010210T000000',
                   {'VALUE': 'PERIOD'})
        self.assertRaises(ICalParseError, vevent._extractListOfDates, 'RDATE',
                                          vevent.rdate_types, True)


class TestReadIcalendar(unittest.TestCase):

    example_ical = dedent("""\
        BEGIN:VCALENDAR
        VERSION
         :2.0
        PRODID
         :-//Mozilla.org/NONSGML Mozilla Calendar V1.0//EN
        METHOD
         :PUBLISH
        BEGIN:VEVENT
        UID
         :956630271
        SUMMARY
         :Christmas Day
        CLASS
         :PUBLIC
        X-MOZILLA-ALARM-DEFAULT-UNITS
         :minutes
        X-MOZILLA-ALARM-DEFAULT-LENGTH
         :15
        X-MOZILLA-RECUR-DEFAULT-UNITS
         :weeks
        X-MOZILLA-RECUR-DEFAULT-INTERVAL
         :1
        DTSTART
         ;VALUE=DATE
         :20031225
        DTEND
         ;VALUE=DATE
         :20031226
        DTSTAMP
         :20020430T114937Z
        END:VEVENT
        END:VCALENDAR
        BEGIN:VCALENDAR
        VERSION
         :2.0
        PRODID
         :-//Mozilla.org/NONSGML Mozilla Calendar V1.0//EN
        METHOD
         :PUBLISH
        BEGIN:VEVENT
        UID
         :911737808
        SUMMARY
         :Boxing Day
        CLASS
         :PUBLIC
        X-MOZILLA-ALARM-DEFAULT-UNITS
         :minutes
        X-MOZILLA-ALARM-DEFAULT-LENGTH
         :15
        X-MOZILLA-RECUR-DEFAULT-UNITS
         :weeks
        X-MOZILLA-RECUR-DEFAULT-INTERVAL
         :1
        DTSTART
         ;VALUE=DATE
         :20030501
        DTSTAMP
         :20020430T114937Z
        END:VEVENT
        BEGIN:VEVENT
        UID
         :wh4t3v3r
        DTSTART;VALUE=DATE:20031225
        SUMMARY:Christmas again!
        END:VEVENT

        END:VCALENDAR
        """)

    ical_with_timezones = dedent("""\
        BEGIN:VCALENDAR
        VERSION:2.0
        X-WR-CALNAME:testas
        PRODID:-//Apple Computer\, Inc//iCal 1.5//EN
        X-WR-RELCALID:42A41871-B382-11DA-BA4E-000A95CA5872
        X-WR-TIMEZONE:Europe/Helsinki
        CALSCALE:GREGORIAN
        METHOD:PUBLISH
        BEGIN:VTIMEZONE
        TZID:Europe/Berlin
        LAST-MODIFIED:20060314T174500Z
        BEGIN:STANDARD
        DTSTART:20051030T010000
        TZOFFSETTO:+0100
        TZOFFSETFROM:+0000
        TZNAME:CET
        END:STANDARD
        BEGIN:DAYLIGHT
        DTSTART:20060326T020000
        TZOFFSETTO:+0200
        TZOFFSETFROM:+0100
        TZNAME:CEST
        END:DAYLIGHT
        END:VTIMEZONE
        BEGIN:VTIMEZONE
        TZID:Asia/Shanghai
        LAST-MODIFIED:20060314T174500Z
        BEGIN:STANDARD
        DTSTART:19321213T204552
        TZOFFSETTO:+0800
        TZOFFSETFROM:+0000
        TZNAME:CST
        END:STANDARD
        BEGIN:DAYLIGHT
        DTSTART:19400603T000000
        TZOFFSETTO:+0900
        TZOFFSETFROM:+0800
        TZNAME:CDT
        END:DAYLIGHT
        BEGIN:STANDARD
        DTSTART:19401001T000000
        TZOFFSETTO:+0800
        TZOFFSETFROM:+0900
        TZNAME:CST
        END:STANDARD
        BEGIN:DAYLIGHT
        DTSTART:19410316T000000
        TZOFFSETTO:+0900
        TZOFFSETFROM:+0800
        TZNAME:CDT
        END:DAYLIGHT
        BEGIN:STANDARD
        DTSTART:19411001T000000
        TZOFFSETTO:+0800
        TZOFFSETFROM:+0900
        TZNAME:CST
        END:STANDARD
        BEGIN:DAYLIGHT
        DTSTART:19860504T000000
        TZOFFSETTO:+0900
        TZOFFSETFROM:+0800
        TZNAME:CDT
        END:DAYLIGHT
        BEGIN:STANDARD
        DTSTART:19860914T000000
        TZOFFSETTO:+0800
        TZOFFSETFROM:+0900
        TZNAME:CST
        END:STANDARD
        BEGIN:DAYLIGHT
        DTSTART:19870412T000000
        TZOFFSETTO:+0900
        TZOFFSETFROM:+0800
        TZNAME:CDT
        END:DAYLIGHT
        BEGIN:STANDARD
        DTSTART:19870913T000000
        TZOFFSETTO:+0800
        TZOFFSETFROM:+0900
        TZNAME:CST
        END:STANDARD
        BEGIN:DAYLIGHT
        DTSTART:19880410T000000
        TZOFFSETTO:+0900
        TZOFFSETFROM:+0800
        TZNAME:CDT
        END:DAYLIGHT
        BEGIN:STANDARD
        DTSTART:19880911T000000
        TZOFFSETTO:+0800
        TZOFFSETFROM:+0900
        TZNAME:CST
        END:STANDARD
        BEGIN:DAYLIGHT
        DTSTART:19890416T000000
        TZOFFSETTO:+0900
        TZOFFSETFROM:+0800
        TZNAME:CDT
        END:DAYLIGHT
        BEGIN:STANDARD
        DTSTART:19890917T000000
        TZOFFSETTO:+0800
        TZOFFSETFROM:+0900
        TZNAME:CST
        END:STANDARD
        BEGIN:DAYLIGHT
        DTSTART:19900415T000000
        TZOFFSETTO:+0900
        TZOFFSETFROM:+0800
        TZNAME:CDT
        END:DAYLIGHT
        BEGIN:STANDARD
        DTSTART:19900916T000000
        TZOFFSETTO:+0800
        TZOFFSETFROM:+0900
        TZNAME:CST
        END:STANDARD
        BEGIN:DAYLIGHT
        DTSTART:19910414T000000
        TZOFFSETTO:+0900
        TZOFFSETFROM:+0800
        TZNAME:CDT
        END:DAYLIGHT
        BEGIN:STANDARD
        DTSTART:19910915T000000
        TZOFFSETTO:+0800
        TZOFFSETFROM:+0900
        TZNAME:CST
        END:STANDARD
        END:VTIMEZONE
        BEGIN:VTIMEZONE
        TZID:Europe/Helsinki
        LAST-MODIFIED:20060314T174500Z
        BEGIN:STANDARD
        DTSTART:20051030T010000
        TZOFFSETTO:+0200
        TZOFFSETFROM:+0000
        TZNAME:EET
        END:STANDARD
        BEGIN:DAYLIGHT
        DTSTART:20060326T040000
        TZOFFSETTO:+0300
        TZOFFSETFROM:+0200
        TZNAME:EEST
        END:DAYLIGHT
        END:VTIMEZONE
        BEGIN:VEVENT
        DTSTART;TZID=Europe/Helsinki:20060307T200000
        SUMMARY:Europe/helsinki
        UID:42A40DBE-B382-11DA-BA4E-000A95CA5872
        SEQUENCE:1
        DTSTAMP:20060314T174340Z
        DURATION:PT1H
        END:VEVENT
        BEGIN:VEVENT
        DTSTART;TZID=Asia/Shanghai:20060308T200000
        SUMMARY:Asia/Shanghai
        UID:42A411C1-B382-11DA-BA4E-000A95CA5872
        SEQUENCE:2
        DTSTAMP:20060314T174418Z
        DURATION:PT1H
        END:VEVENT
        BEGIN:VEVENT
        DTSTART;TZID=Europe/Berlin:20060309T200000
        SUMMARY:Europe/Berlin
        UID:42A4138F-B382-11DA-BA4E-000A95CA5872
        SEQUENCE:2
        DTSTAMP:20060314T174433Z
        DURATION:PT1H
        END:VEVENT
        END:VCALENDAR
        """)

    def doctest_error_handling(self):
        r"""Test for read_icalendar error handling.

        Let's try parsing some invalid icalendar files:

            >>> from schooltool.calendar.icalendar import read_icalendar

            >>> file = StringIO(dedent('''\
            ...             BEGIN:VCALENDAR
            ...             BEGIN:VEVENT
            ...             DTSTART;VALUE=DATE:20010203
            ...             END:VEVENT
            ...             END:VCALENDAR
            ...             '''))
            >>> list(read_icalendar(file))
            Traceback (most recent call last):
            ...
            ICalParseError: VEVENT must have a UID property

            >>> file = StringIO(dedent('''\
            ...             BEGIN:VCALENDAR
            ...             BEGIN:VEVENT
            ...             UID:hello
            ...             END:VEVENT
            ...             END:VCALENDAR
            ...             '''))
            >>> list(read_icalendar(file))
            Traceback (most recent call last):
            ...
            ICalParseError: VEVENT must have a DTSTART property

            >>> file = StringIO(dedent('''\
            ...             BEGIN:VCALENDAR
            ...             BEGIN:VEVENT
            ...             UID:hello
            ...             DTSTART;VALUE=DATUM:20010203T101516
            ...             END:VEVENT
            ...             END:VCALENDAR
            ...             '''))
            >>> list(read_icalendar(file))
            Traceback (most recent call last):
            ...
            ICalParseError: DTSTART property should have a DATE or DATE-TIME value

            >>> file = StringIO(dedent('''\
            ...             BEGIN:VCALENDAR
            ...             BEGIN:VEVENT
            ...             UID:hello
            ...             DTSTART;VALUE=DATE-TIME:20010203T101516
            ...             DTEND;VALUE=DATE:20010203
            ...             END:VEVENT
            ...             END:VCALENDAR
            ...             '''))
            >>> list(read_icalendar(file))
            Traceback (most recent call last):
            ...
            ICalParseError: DTEND property should have the same type as DTSTART

            >>> file = StringIO(dedent('''\
            ...             BEGIN:VCALENDAR
            ...             BEGIN:VEVENT
            ...             UID:hello
            ...             DTSTART;VALUE=DATE:20010201
            ...             DURATION;VALUE=DAYS:P1DT1H1M2S
            ...             END:VEVENT
            ...             END:VCALENDAR
            ...             '''))
            >>> list(read_icalendar(file))
            Traceback (most recent call last):
            ...
            ICalParseError: DURATION property should have type DURATION

            >>> file = StringIO(dedent('''\
            ...             BEGIN:VCALENDAR
            ...             BEGIN:VEVENT
            ...             DTSTART;VALUE=DATE:20010203
            ...             '''))
            >>> list(read_icalendar(file))
            Traceback (most recent call last):
            ...
            ICalParseError: Mismatched BEGIN/END

            >>> file = StringIO(dedent('''\
            ...             BEGIN:VCALENDAR
            ...             BEGIN:VEVENT
            ...             DTSTART;VALUE=DATE:20010203
            ...             END:VCALENDAR
            ...             END:VEVENT
            ...             '''))
            >>> list(read_icalendar(file))
            Traceback (most recent call last):
            ...
            ICalParseError: Mismatched BEGIN/END


            >>> file = StringIO(dedent('''\
            ...             BEGIN:VEVENT
            ...             END:VEVENT
            ...             '''))
            >>> list(read_icalendar(file))
            Traceback (most recent call last):
            ...
            ICalParseError: This is not iCalendar

            >>> file = StringIO(dedent('''\
            ...             BEGIN:VCALENDAR
            ...             END:VCALENDAR
            ...             X-PROP:foo
            ...             '''))
            >>> list(read_icalendar(file))
            Traceback (most recent call last):
            ...
            ICalParseError: Text outside VCALENDAR component

            >>> file = StringIO(dedent('''\
            ...             BEGIN:VCALENDAR
            ...             END:VCALENDAR
            ...             BEGIN:VEVENT
            ...             END:VEVENT
            ...             '''))
            >>> list(read_icalendar(file))
            Traceback (most recent call last):
            ...
            ICalParseError: Text outside VCALENDAR component

            >>> file = StringIO(dedent('''\
            ...             BEGIN:VCALENDAR
            ...             BEGIN:VEVENT
            ...             DTSTART;VALUE=DATE:20010203
            ...             END:VEVENT
            ...             END:VCALENDAR
            ...             END:UNIVERSE
            ...             '''))
            >>> list(read_icalendar(file))
            Traceback (most recent call last):
            ...
            ICalParseError: Mismatched BEGIN/END

            >>> file = StringIO(dedent('''\
            ...             DTSTART;VALUE=DATE:20010203
            ...             '''))
            >>> list(read_icalendar(file))
            Traceback (most recent call last):
            ...
            ICalParseError: Text outside VCALENDAR component

            >>> file = StringIO(dedent('''\
            ...             This is just plain text
            ...             '''))
            >>> list(read_icalendar(file))
            Traceback (most recent call last):
            ...
            ICalParseError: Expected ':' in line:
            This is just plain text

        If you pass both DTEND and DURATION, no errors are raised:

            >>> file = StringIO(dedent('''\
            ...             BEGIN:VCALENDAR
            ...             BEGIN:VEVENT
            ...             UID:hello
            ...             DTSTART;VALUE=DATE:20010201
            ...             DTEND;VALUE=DATE:20010203
            ...             DURATION:P1D
            ...             END:VEVENT
            ...             END:VCALENDAR
            ...             '''))
            >>> event = list(read_icalendar(file))[0]

        The DURATION is used, while DTEND gets ignored:

            >>> event.duration
            datetime.timedelta(1)

        """

    def test_read_icalendar_with_timezones(self):
        from schooltool.calendar.icalendar import read_icalendar
        from pytz import utc

        file = StringIO(self.ical_with_timezones)
        result = list(read_icalendar(file))
        self.assertEqual(len(result), 3)
        event = result[0]

        self.assertEqual(event.dtstart, datetime(2006, 3, 7, 18, 0, tzinfo=utc))

        event = result[1]
        self.assertEqual(event.dtstart, datetime(2006, 3, 8, 12, 0, tzinfo=utc))

        event = result[2]
        self.assertEqual(event.dtstart, datetime(2006, 3, 9, 19, 0, tzinfo=utc))

    def test_read_icalendar(self):
        from schooltool.calendar.icalendar import read_icalendar
        from pytz import utc
        file = StringIO(self.example_ical)
        result = list(read_icalendar(file))
        self.assertEqual(len(result), 3)
        event = result[0]


        self.assertEqual(event.dtstart, datetime(2003, 12, 25, 0, 0, tzinfo=utc))
        self.assertEqual(event.duration, timedelta(days=1))
        event = result[1]
        self.assertEqual(event.dtstart, datetime(2003, 5, 1, 0, 0, tzinfo=utc))
        event = result[2]
        self.assertEqual(event.duration, timedelta(days=1))

        result = list(read_icalendar(StringIO(dedent("""\
                    BEGIN:VCALENDAR

                    BEGIN:VEVENT
                    UID:hello
                    DTSTART;VALUE=DATE:20010203
                    BEGIN:VALARM
                    X-PROP:foo
                    END:VALARM
                    END:VEVENT

                    END:VCALENDAR
                    """))))
        self.assertEquals(len(result), 1)
        event = result[0]
        self.assertEquals(event.unique_id, 'hello')
        self.assert_(event.dtstart)

        file = StringIO("")
        self.assertEquals(list(read_icalendar(file)), [])


def doctest_read_icalendar_floating():
    r"""Tests for reading of events with floating time.

        >>> from schooltool.calendar.icalendar import read_icalendar
        >>> calendar = dedent('''\
        ...     BEGIN:VCALENDAR
        ...     BEGIN:VEVENT
        ...     UID:hello
        ...     DTSTART;VALUE=DATE-TIME:20010203T101516
        ...     BEGIN:VALARM
        ...     X-PROP:foo
        ...     END:VALARM
        ...     END:VEVENT
        ...
        ...     END:VCALENDAR''')

    If no fallback_tz is passed, floating time events are converted to
    UTC:

        >>> event = list(read_icalendar(calendar))[0]
        >>> event.dtstart
        datetime.datetime(2001, 2, 3, 10, 15, 16, tzinfo=<UTC>)

    But if we pass a fallback_tz it should be used:

        >>> import pytz
        >>> tzinfo = pytz.timezone('Europe/Vilnius')
        >>> event = list(read_icalendar(calendar, fallback_tz=tzinfo))[0]
        >>> event.dtstart
        datetime.datetime(2001, 2, 3, 8, 15, 16, tzinfo=<UTC>)

    """


def doctest_read_icalendar_unrecognized_timezone():
    r"""Tests for reading of events with unrecognized timezone.

    If event has a timezone we can't recognize set:

        >>> from schooltool.calendar.icalendar import read_icalendar
        >>> calendar = dedent('''\
        ...     BEGIN:VCALENDAR
        ...
        ...     BEGIN:VTIMEZONE
        ...     TZID:Yurop/Berlin
        ...     BEGIN:STANDARD
        ...     TZNAME:YNAME
        ...     END:STANDARD
        ...     END:VTIMEZONE
        ...
        ...     BEGIN:VEVENT
        ...     UID:some-random-uid@example.com
        ...     DTSTART;TZID=Yurop/Berlin:20050226T160000
        ...     DURATION:PT6H
        ...     DTSTAMP:20050203T150000
        ...     END:VEVENT
        ...
        ...     END:VCALENDAR''')

    they are treated as if they were happening in UTC:

        >>> event = list(read_icalendar(calendar))[0]
        >>> event.dtstart
        datetime.datetime(2005, 2, 26, 16, 0, tzinfo=<UTC>)

    But if we pass a fallback_tz it should be used:

        >>> import pytz
        >>> tzinfo = pytz.timezone('Europe/Vilnius')
        >>> event = list(read_icalendar(calendar, fallback_tz=tzinfo))[0]
        >>> event.dtstart
        datetime.datetime(2005, 2, 26, 14, 0, tzinfo=<UTC>)

    """


def doctest_read_icalendar_allday_dtend():
    r"""Tests for reading of allday events with DTEND.

    Duration of all-day events should get properly calculated:

        >>> from schooltool.calendar.icalendar import read_icalendar
        >>> calendar = dedent('''\
        ...     BEGIN:VCALENDAR
        ...
        ...     BEGIN:VEVENT
        ...     UID:some-random-uid@example.com
        ...     DTSTART;VALUE=DATE:20050226
        ...     DTEND;VALUE=DATE:20050228
        ...     DTSTAMP:20050203T150000
        ...     END:VEVENT
        ...
        ...     END:VCALENDAR''')

        >>> event = list(read_icalendar(calendar))[0]
        >>> event.dtstart
        datetime.datetime(2005, 2, 26, 0, 0, tzinfo=<UTC>)
        >>> event.duration
        datetime.timedelta(2)

    """


class TestRowParser(unittest.TestCase):

    def test_parse(self):
        from schooltool.calendar.icalendar import RowParser
        file = StringIO("key1\n"
                        " :value1\n"
                        " \n"
                        "key2\n"
                        " ;VALUE=foo\n"
                        " :value2\n"
                        "key3;VALUE=bar:value3\n")
        self.assertEqual(list(RowParser.parse(file)),
                         [('KEY1', 'value1', {}),
                          ('KEY2', 'value2', {'VALUE': 'FOO'}),
                          ('KEY3', 'value3', {'VALUE': 'BAR'})])

        file = StringIO("key1:value1\n"
                        "key2;VALUE=foo:value2\n"
                        "key3;VALUE=bar:value3\n")
        self.assertEqual(list(RowParser.parse(file)),
                         [('KEY1', 'value1', {}),
                          ('KEY2', 'value2', {'VALUE': 'FOO'}),
                          ('KEY3', 'value3', {'VALUE': 'BAR'})])

        file = StringIO("key1:value:with:colons:in:it\n")
        self.assertEqual(list(RowParser.parse(file)),
                         [('KEY1', 'value:with:colons:in:it', {})])

        file = StringIO("ke\r\n y1\n\t:value\r\n  1 \r\n .")
        self.assertEqual(list(RowParser.parse(file)),
                         [('KEY1', 'value 1 .', {})])

        file = StringIO("key;param=\xe2\x98\xbb:\r\n"
                        " value \xe2\x98\xbb\r\n")
        self.assertEqual(list(RowParser.parse(file)),
                         [("KEY", u"value \u263B", {'PARAM': u'\u263B'})])

    def test_parseRow(self):
        from schooltool.calendar.icalendar import RowParser, ICalParseError
        parseRow = RowParser._parseRow
        self.assertEqual(parseRow("key:"), ("KEY", "", {}))
        self.assertEqual(parseRow("key:value"), ("KEY", "value", {}))
        self.assertEqual(parseRow("key:va:lu:e"), ("KEY", "va:lu:e", {}))
        self.assertRaises(ICalParseError, parseRow, "key but no value")
        self.assertRaises(ICalParseError, parseRow, ":value but no key")
        self.assertRaises(ICalParseError, parseRow, "bad name:")

        self.assertEqual(parseRow("key;param=:value"),
                         ("KEY", "value", {'PARAM': ''}))
        self.assertEqual(parseRow("key;param=pvalue:value"),
                         ("KEY", "value", {'PARAM': 'PVALUE'}))
        self.assertEqual(parseRow('key;param=pvalue;param2=value2:value'),
                         ("KEY", "value", {'PARAM': 'PVALUE',
                                           'PARAM2': 'VALUE2'}))
        self.assertEqual(parseRow('key;param="pvalue":value'),
                         ("KEY", "value", {'PARAM': 'pvalue'}))
        self.assertEqual(parseRow('key;param=pvalue;param2="value2":value'),
                         ("KEY", "value", {'PARAM': 'PVALUE',
                                           'PARAM2': 'value2'}))
        self.assertRaises(ICalParseError, parseRow, "k;:no param")
        self.assertRaises(ICalParseError, parseRow, "k;a?=b:bad param")
        self.assertRaises(ICalParseError, parseRow, "k;a=\":bad param")
        self.assertRaises(ICalParseError, parseRow, "k;a=\"\177:bad param")
        self.assertRaises(ICalParseError, parseRow, "k;a=\001:bad char")
        self.assertEqual(parseRow("key;param=a,b,c:value"),
                         ("KEY", "value", {'PARAM': ['A', 'B', 'C']}))
        self.assertEqual(parseRow('key;param=a,"b,c",d:value'),
                         ("KEY", "value", {'PARAM': ['A', 'b,c', 'D']}))


def doctest_VTimezone():
    r"""Test for VTimezone.

        >>> from schooltool.calendar.icalendar import VTimezone, RowParser
        >>> tz = VTimezone('SchoolTool-Europe/Vilnius', ['foo'])
        >>> tz.tzid
        'SchoolTool-Europe/Vilnius'
        >>> tz.tznames
        ['foo']

        >>> example_vtimezone = dedent('''
        ... BEGIN:VTIMEZONE
        ... TZID:Europe/Berlin
        ... LAST-MODIFIED:20060314T174500Z
        ... BEGIN:STANDARD
        ... DTSTART:20051030T010000
        ... TZOFFSETTO:+0100
        ... TZOFFSETFROM:+0000
        ... TZNAME:CET
        ... END:STANDARD
        ... BEGIN:DAYLIGHT
        ... DTSTART:20060326T020000
        ... TZOFFSETTO:+0200
        ... TZOFFSETFROM:+0100
        ... TZNAME:CEST
        ... END:DAYLIGHT
        ... END:VTIMEZONE
        ... ''')

        >>> rows = RowParser.parse(example_vtimezone.splitlines())
        >>> vtz = VTimezone.parse(rows)

        >>> vtz.tzid
        'Europe/Berlin'
        >>> vtz.tznames
        [u'CET']

    In Evolution calendars X-LIC-LOCATION is very useful for mapping to
    actual pytz timezones, so we store it too if we find it:

        >>> example_vtimezone = dedent('''
        ... BEGIN:VTIMEZONE
        ... TZID:Evolution-blahblah@@obscure_Europe/Berlin
        ... X-LIC-LOCATION:Europe/Berlin
        ... LAST-MODIFIED:20060314T174500Z
        ... BEGIN:STANDARD
        ... DTSTART:20051030T010000
        ... TZOFFSETTO:+0100
        ... TZOFFSETFROM:+0000
        ... TZNAME:CET
        ... END:STANDARD
        ... END:VTIMEZONE
        ... ''')
        >>> rows = RowParser.parse(example_vtimezone.splitlines())
        >>> vtz = VTimezone.parse(rows)
        >>> vtz.tzid
        'Evolution-blahblah@@obscure_Europe/Berlin'
        >>> vtz.x_lic_location
        u'Europe/Berlin'

    """


def doctest_VTimezone_getTzinfo():
    """Tests for VTimezone.getTzinfo()

        >>> from schooltool.calendar.icalendar import VTimezone

    If tzid has a matching timezone, we use it, ignoring other
    arguments:

        >>> vtimezone = VTimezone("Europe/Berlin",
        ...                       ['CET', 'CEST'],
        ...                       x_lic_location="Europe/Vilnius")
        >>> vtimezone.getTzinfo()
        <DstTzInfo 'Europe/Berlin' CET+1:00:00 STD>

    If not, we iterate through tzname list and return the first match:

        >>> vtimezone = VTimezone("Evolution-blahblah@@obscure_Europe/Berlin",
        ...                       ['CET', 'CEST'])
        >>> vtimezone.getTzinfo()
        <DstTzInfo 'CET' CET+1:00:00 STD>

    We iterate from end to front, because in most ics files more
    recent tznames are in the end of the list:

        >>> vtimezone = VTimezone("Evolution-blahblah@@obscure_Europe/Berlin",
        ...                       ['CET', 'EET'])
        >>> vtimezone.getTzinfo()
        <DstTzInfo 'EET' EET+2:00:00 STD>

    If x_lic_location is supplied and has a matching timezone, tznames
    are ignored. x_lic_location is used only if tzid has no matching
    timezone:

        >>> vtimezone = VTimezone("Evolution-blahblah@@obscure_Europe/Berlin",
        ...                       ['CET', 'CEST'],
        ...                       x_lic_location="Europe/Vilnius")
        >>> vtimezone.getTzinfo()
        <DstTzInfo 'Europe/Vilnius' WMT+1:24:00 STD>

    """


def doctest_VTimezone_errors():
    r"""Test for VTimezone error handling.

        >>> from schooltool.calendar.icalendar import VTimezone, RowParser

    First, let's omit the timezone ID:

        >>> example_vtimezone = dedent('''
        ... BEGIN:VTIMEZONE
        ... LAST-MODIFIED:20060314T174500Z
        ... BEGIN:STANDARD
        ... DTSTART:20051030T010000
        ... TZOFFSETTO:+0100
        ... TZOFFSETFROM:+0000
        ... TZNAME:CET
        ... END:STANDARD
        ... END:VTIMEZONE
        ... ''')
        >>> rows = RowParser.parse(example_vtimezone.splitlines())
        >>> VTimezone.parse(rows)
        Traceback (most recent call last):
            ...
        ICalParseError: Missing TZID in VTIMEZONE block

    Then omit the STANDARD section:

        >>> example_vtimezone = dedent('''
        ... BEGIN:VTIMEZONE
        ... TZID:Europe/Berlin
        ... LAST-MODIFIED:20060314T174500Z
        ... END:VTIMEZONE
        ... ''')
        >>> rows = RowParser.parse(example_vtimezone.splitlines())
        >>> VTimezone.parse(rows)
        Traceback (most recent call last):
            ...
        ICalParseError: Missing STANDARD section in VTIMEZONE block

    """


def doctest_VCalendar():
    """Test for VCalendar.

        >>> from schooltool.calendar.icalendar import VCalendar, RowParser
        >>> vcal = VCalendar(['event1'], ['timezone2'])
        >>> vcal.events
        ['event1']
        >>> vcal.timezones
        ['timezone2']

    Now we will check the parser

        >>> example_ical = dedent('''
        ... BEGIN:VCALENDAR
        ... VERSION:2.0
        ... PRODID:-//SchoolTool.org/NONSGML SchoolTool//EN
        ...
        ... BEGIN:VTIMEZONE
        ... TZID:Europe/Berlin
        ... BEGIN:STANDARD
        ... TZNAME:CET
        ... END:STANDARD
        ... END:VTIMEZONE
        ...
        ... BEGIN:VEVENT
        ... UID:some-random-uid@example.com
        ... DTSTART:20050226T160000
        ... DURATION:PT6H
        ... DTSTAMP:20050203T150000
        ... END:VEVENT
        ... END:VCALENDAR
        ... ''')
        >>> rows = list(RowParser.parse(example_ical.splitlines()))
        >>> vcal = VCalendar.parse(rows)
        >>> pprint(vcal.timezones)
        {'EUROPE/BERLIN': <DstTzInfo 'Europe/Berlin' CET+1:00:00 STD>,
         'UTC': <UTC>}
        >>> vcal.events
        [<schooltool.calendar.icalendar.VEvent object at ...>]

    If there is a vtimezone block that we can't find a tzinfo for we
    are skipping it:

        >>> example_ical = dedent('''
        ... BEGIN:VCALENDAR
        ... VERSION:2.0
        ... PRODID:-//SchoolTool.org/NONSGML SchoolTool//EN
        ...
        ... BEGIN:VTIMEZONE
        ... TZID:Europe/Berlin
        ... BEGIN:STANDARD
        ... TZNAME:CET
        ... END:STANDARD
        ... END:VTIMEZONE
        ...
        ... BEGIN:VTIMEZONE
        ... TZID:Yurop/Berlin
        ... BEGIN:STANDARD
        ... TZNAME:YNAME
        ... END:STANDARD
        ... END:VTIMEZONE
        ...
        ... BEGIN:VEVENT
        ... UID:some-random-uid@example.com
        ... DTSTART:20050226T160000
        ... DURATION:PT6H
        ... DTSTAMP:20050203T150000
        ... END:VEVENT
        ... END:VCALENDAR
        ... ''')
        >>> rows = list(RowParser.parse(example_ical.splitlines()))
        >>> vcal = VCalendar.parse(rows)
        >>> pprint(vcal.timezones)
        {'EUROPE/BERLIN': <DstTzInfo 'Europe/Berlin' CET+1:00:00 STD>,
         'UTC': <UTC>}
        >>> vcal.events
        [<schooltool.calendar.icalendar.VEvent object at ...>]

    """


def doctest_ical_reader_empty_summary():
    r"""Regression test for read_icalendar

    Mozilla Calendar allows events with an empty summary.  This used to be
    read as a CalendarEvent with title = None in schooltool, which broke
    things.

        >>> from schooltool.calendar.icalendar import read_icalendar
        >>> events = list(read_icalendar('''\
        ... BEGIN:VCALENDAR
        ... VERSION:2.0
        ... PRODID:-//SchoolTool.org/NONSGML SchoolTool//EN
        ... BEGIN:VEVENT
        ... UID:some-random-uid@example.com
        ... DTSTART:20050226T160000
        ... DURATION:PT6H
        ... DTSTAMP:20050203T150000
        ... END:VEVENT
        ... END:VCALENDAR
        ... '''))
        >>> [e.title for e in events]
        ['']

    We support reading calendars when ical files are using different
    charset too:

        >>> events = list(read_icalendar('''\
        ... BEGIN:VCALENDAR
        ... VERSION:2.0
        ... PRODID:-//SchoolTool.org/NONSGML SchoolTool//EN
        ... BEGIN:VEVENT
        ... UID:some-random-uid@example.com
        ... SUMMARY:LAN party %s
        ... DTSTART:20050226T160000
        ... DURATION:PT6H
        ... DTSTAMP:20050203T150000
        ... END:VEVENT
        ... END:VCALENDAR
        ... ''' %  chr(163), charset='latin-1'))
        >>> titles = [e.title for e in events]
        >>> titles[0]
        u'LAN party \xa3'

    """


def doctest_read_icalendar_all_day():
    r"""Test for generation of allday events.

    SchoolTool should import all day events as such:

        >>> from schooltool.calendar.icalendar import read_icalendar
        >>> events = list(read_icalendar('''\
        ... BEGIN:VCALENDAR
        ... VERSION:2.0
        ... PRODID:-//SchoolTool.org/NONSGML SchoolTool//EN
        ... BEGIN:VEVENT
        ... SUMMARY:All day event
        ... UID:some-random-uid@example.com
        ... DTSTART;VALUE=DATE:20050226
        ... DURATION:P2D
        ... DTSTAMP:20050203T150000
        ... END:VEVENT
        ... END:VCALENDAR
        ... '''))
        >>> len(events)
        1

        >>> event = list(events)[0]
        >>> event.title
        u'All day event'
        >>> event.duration
        datetime.timedelta(2)
        >>> event.allday
        True
        >>> event.dtstart
        datetime.datetime(2005, 2, 26, 0, 0, tzinfo=<UTC>)

    """


def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(doctest.DocTestSuite(
                        optionflags=doctest.ELLIPSIS | doctest.REPORT_UDIFF |
                                    doctest.NORMALIZE_WHITESPACE))
    suite.addTest(doctest.DocTestSuite('schooltool.calendar.icalendar',
                        optionflags=doctest.ELLIPSIS | doctest.REPORT_UDIFF |
                                    doctest.NORMALIZE_WHITESPACE))
    suite.addTest(unittest.makeSuite(TestParseDateTime))
    suite.addTest(unittest.makeSuite(TestPeriod))
    suite.addTest(unittest.makeSuite(TestVEvent))
    suite.addTest(unittest.makeSuite(TestReadIcalendar))
    suite.addTest(unittest.makeSuite(TestRowParser))
    return suite

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2012 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Tests for SchoolTool catalogs.
"""
import unittest
import doctest

from schooltool.app.catalog import buildQueryString


def doctest_buildQueryString():
    r"""Test buildQueryString

    No text yields empty string.

        >>> buildQueryString('')
        ''

        >>> buildQueryString(' \t \n ')
        ''

    Words are globbed.  Text is lowercase.

        >>> buildQueryString(' Tim Alan ')
        'tim* alan*'

    Commas are replaced by 'or' term.

        >>> buildQueryString(' Tim,   Alan ')
        'tim* or alan*'

        >>> buildQueryString(', Tim, ,,,  Alan ,')
        'tim* or alan*'

    Broken testcases follow.

        >>> buildQueryString('"Timothy" Alan')
        '"timothy" alan*'

        >>> buildQueryString('"Timothy, Alan"')
        '"timothy, alan"'

    Mister Tim or Tom, last name Cook or Cake:

        >>> buildQueryString('T?m C*k*')
        't?m* c*k*'

    Mister Or Arwell:

        >>> buildQueryString('"Or" Arwell')
        '"or" arwell*'

    """


def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(doctest.DocTestSuite(
        optionflags=doctest.ELLIPSIS|doctest.REPORT_NDIFF|
                    doctest.NORMALIZE_WHITESPACE))
    return suite


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Testing Utilities
"""
__metaclass__ = type
import cgi
import difflib
import unittest
from lxml import sax
from lxml import etree
from xml.sax.handler import ContentHandler
from pprint import pformat
from StringIO import StringIO

from zope.interface import implements
from zope.traversing.interfaces import IContainmentRoot


def unidiff(old, new, oldlabel="expected output", newlabel="actual output"):
    """Display a compact unified diff between old text and new text.

    Bugs: does not work when old or new is an empty string (this seems to
    be a bug/limitation of the Python 2.3 difflib module).
    """
    return "\n".join(difflib.unified_diff(old.splitlines(), new.splitlines(),
                                          oldlabel, newlabel, lineterm=''))


def diff(old, new, oldlabel="expected output", newlabel="actual output"):
    """Display a unified diff between old text and new text."""
    old = old.splitlines()
    new = new.splitlines()

    diff = ['--- %s' % oldlabel, '+++ %s' % newlabel]

    def dump(tag, x, lo, hi):
        for i in xrange(lo, hi):
            diff.append(tag + x[i])

    differ = difflib.SequenceMatcher(a=old, b=new)
    for tag, alo, ahi, blo, bhi in differ.get_opcodes():
        if tag == 'replace':
            dump('-', old, alo, ahi)
            dump('+', new, blo, bhi)
        elif tag == 'delete':
            dump('-', old, alo, ahi)
        elif tag == 'insert':
            dump('+', new, blo, bhi)
        elif tag == 'equal':
            dump(' ', old, alo, ahi)
        else:
            raise AssertionError('unknown tag %r' % tag)
    return "\n".join(diff)


def format_table(table, header_rows=0):
    """Format an ASCII-art table.

    Example:

      >>> print format_table([['11', '12', '13'],
      ...                     ['21', 'center', '23'],
      ...                     ['31', 32, None]])
      +----+--------+------+
      | 11 | 12     | 13   |
      | 21 | center | 23   |
      | 31 | 32     | None |
      +----+--------+------+

      >>> print format_table([])
      +-+
      +-+

      >>> print format_table([['', 'x']])
      +-+---+
      | | x |
      +-+---+

      >>> print format_table([['x', 'y', 'z'],
      ...                     ['11', '12', '13'],
      ...                     ['31', 32, '33']], header_rows=1)
      +----+----+----+
      | x  | y  | z  |
      +----+----+----+
      | 11 | 12 | 13 |
      | 31 | 32 | 33 |
      +----+----+----+

    """
    ncols = table and len(table[0]) or 1
    col_width = [1] * ncols
    for row_data in table:
        for col, cell_data in enumerate(row_data):
            if cell_data != '':
                col_width[col] = max(col_width[col], len(str(cell_data)) + 2)
    hline = '+'.join([''] + ['-' * w for w in col_width] + [''])
    table_rows = ([hline] +
                  ['|'.join([''] +
                            [' ' + str(s).ljust(w - 1)
                             for s, w in zip(row_data, col_width)] +
                            [''])
                   for row_data in table] +
                  [hline])
    if 0 < header_rows < len(table):
        table_rows.insert(1 + header_rows, hline)
    table = '\n'.join(table_rows)
    return table


def pformat_set(s):
    """Pretty-print a Set."""
    items = list(s)
    items.sort()
    return 'set(%s)' % pformat(items)



def normalize_xml(xml, recursively_sort=(), compact=False):
    """Normalizes an XML document.

    The idea is that two semantically equivalent XML documents should be
    normalized into the same canonical representation.  Therefore if two
    documents compare equal after normalization, they are semantically
    equivalent.

    The canonical representation used here has nothing to do with W3C Canonical
    XML.

    This function normalizes indentation, whitespace and newlines (except
    inside text nodes), element attribute order, expands character references,
    expands shorthand notation of empty XML elements ("<br/>" becomes
    "<br></br>").

    If recursively_sort is given, it is a sequence of tags that will have
    test:sort="recursively" automatically appended to their attribute lists in
    the text.  Use it when you cannot or do not want to modify the XML document
    itself.

    If compact is True, nodes that only have text (without newlines) will be
    presented more compactly ("<tag>text</tag>").

    Caveats:
     - normalize_xml does not deal well with text nodes
     - normalize_xml does not help when different prefixes are used for the
       same namespace
     - normalize_xml does not handle all XML features (CDATA sections, inline
       DTDs, processing instructions, comments)
    """

    class Document:

        def __init__(self):
            self.children = []
            self.sort_recursively = False

        def render(self, level=0):
            result = []
            for child in self.children:
                result.append(child.render(level))
            return ''.join(result)

    class Element:

        def __init__(self, parent, tag, attrlist, sort=False,
                     sort_recursively=False):
            self.parent = parent
            self.tag = tag
            self.attrlist = attrlist
            self.children = []
            self.sort = sort
            self.sort_recursively = sort_recursively

        def render(self, level):
            result = []
            indent = '  ' * level
            line = '%s<%s' % (indent, self.tag[1])
            for attr in self.attrlist:
                if len(line + attr) < 78:
                    line += attr
                else:
                    result.append(line)
                    result.append('\n')
                    line = '%s %s%s' % (indent, ' ' * len(self.tag[1]), attr)
            if self.children:
                s = ''.join([child.render(level+1) for child in self.children])
            else:
                s = ''
            if not s:
                result.append('%s/>\n' % line)
            elif (compact and len(self.children) == 1 and '<' not in s
                  and s.count('\n') == 1):
                result.append('%s>%s</%s>\n' % (line, s.strip(), self.tag[1]))
            else:
                result.append('%s>\n' % line)
                result.append(s)
                result.append('%s</%s>\n' % (indent, self.tag[1]))
            return ''.join(result)

        def finalize(self):
            if self.sort:
                self.children.sort(lambda x, y: cmp(x.key, y.key))
            self.key = self.render(0)

    class Text:

        def __init__(self, data):
            self.data = data
            self.key = None

        def render(self, level):
            data = cgi.escape(self.data.strip())
            if data:
                indent = '  ' * level
                return ''.join(['%s%s\n' % (indent, line.strip())
                                for line in data.splitlines()])
            else:
                return ''

    class Handler(ContentHandler):

        def __init__(self):
            self.level = 0
            self.result = []
            self.root = self.cur = Document()
            self.last_text = None
            self._locator = None

        def startElementNS(self, tag, qname, attrs):
            self.startElement(tag, attrs)

        def endElementNS(self, tag, qname):
            self.endElement(tag)

        def startElement(self, tag, attrs):
            sort = sort_recursively = self.cur.sort_recursively
            if attrs:
                if tag in recursively_sort:
                    sort = sort_recursively = True
                attrlist = attrs.items()
                attrlist.sort()
                attrlist = [' %s="%s"' % (k[1], cgi.escape(v, True))
                            for k, v in attrlist]
            else:
                attrlist = []
            child = Element(self.cur, tag, attrlist, sort=sort,
                            sort_recursively=sort_recursively)
            self.cur.children.append(child)
            self.cur = child
            self.last_text = None

        def endElement(self, tag):
            self.cur.finalize()
            self.cur = self.cur.parent
            self.last_text = None

        def characters(self, data):
            if self.last_text is not None:
                self.last_text.data += data
            else:
                self.last_text = Text(data)
                self.cur.children.append(self.last_text)

        def render(self):
            return self.root.render()

    for tag in recursively_sort:
        xml = xml.replace('<%s' % tag,
                          '<%s test:sort="recursively"' % tag)
    handler = Handler()
    tree = etree.XML(xml)
    sax.saxify(tree, handler)
    return ''.join(handler.render())


def run_unit_tests(testcase):
    r"""Hack to call into unittest from doctests.

        >>> class SampleTestCase(unittest.TestCase):
        ...     def test1(self):
        ...         self.assertEquals(2 + 2, 4)
        >>> run_unit_tests(SampleTestCase)

        >>> class BadTestCase(SampleTestCase):
        ...     def test2(self):
        ...         self.assertEquals(2 * 2, 5)
        >>> run_unit_tests(BadTestCase) # doctest: +REPORT_NDIFF
        .F
        ======================================================================
        FAIL: test2 (schooltool.app.tests.test_app.BadTestCase)
        ----------------------------------------------------------------------
        Traceback (most recent call last):
        ...
        AssertionError: 4 != 5
        <BLANKLINE>
        ----------------------------------------------------------------------
        Ran 2 tests in ...s
        <BLANKLINE>
        FAILED (failures=1)

    """
    testsuite = unittest.makeSuite(testcase)
    output = StringIO()
    result = unittest.TextTestRunner(output).run(testsuite)
    if not result.wasSuccessful():
        print output.getvalue(),



class _Anything:
    """An object that is equal to any other object."""

    def __eq__(self, other):
        return True

    def __ne__(self, other):
        return False

    def __repr__(self):
        return 'Anything'

Anything = _Anything()


class EqualsSortedMixin:
    """Mixin that adds a helper method for comparing lists ignoring order."""

    def assertEqualsSorted(self, a, b):
        x = list(a)
        y = list(b)
        x.sort()
        y.sort()
        self.assertEquals(x, y)

    assertEqualSorted = assertEqualsSorted


class NiceDiffsMixin:
    """Mixin that changes assertEquals to show a unified diff of pretty-printed
    values.
    """

    def assertEquals(self, results, expected, msg=None):
        if msg is None:
            if (isinstance(expected, basestring)
                and isinstance(results, basestring)):
                msg = "\n" + diff(expected, results)
            elif (isinstance(expected, set)
                and isinstance(results, set)):
                msg = "\n" + diff(pformat_set(expected), pformat_set(results))
            else:
                msg = "\n" + diff(pformat(expected), pformat(results))
        unittest.TestCase.assertEquals(self, results, expected, msg)

    assertEqual = assertEquals



class XMLCompareMixin:

    def assertEqualsXML(self, result, expected, recursively_sort=()):
        """Assert that two XML documents are equivalent.

        If recursively_sort is given, it is a sequence of tags that
        will have test:sort="recursively" appended to their attribute lists
        in 'result' text.  See the docstring for normalize_xml for more
        information about this attribute.
        """
        result = normalize_xml(result, recursively_sort=recursively_sort)
        expected = normalize_xml(expected, recursively_sort=recursively_sort)
        self.assertEquals(result, expected, "\n" + diff(expected, result))

    assertEqualXML = assertEqualsXML


def compareXML(result, expected, recursively_sort=()):
    """Compare 2 XML snippets for equality.

    This is a doctest version of XMLCompareMixin.assertEqualsXML.

    If recursively_sort is given, it is a sequence of tags that will have
    test:sort="recursively" appended to their attribute lists in 'result' text.
    See the docstring for normalize_xml for more information about this
    attribute.
    """
    result = normalize_xml(result, recursively_sort=recursively_sort)
    expected = normalize_xml(expected, recursively_sort=recursively_sort)
    if result == expected:
        return True
    else:
        print diff(expected, result)
        return False


class FakeRoot(object):
    implements(IContainmentRoot)


class FakeFolder(object):
    def __init__(self, parent, name):
        self.__parent__ = parent
        self.__name__ = name


def fakePath(obj, path):
    """Make absoluteURL(obj) return url.

        >>> from zope.traversing.browser.absoluteurl import absoluteURL
        >>> from zope.publisher.browser import TestRequest
        >>> from zope.app.testing import setup
        >>> setup.placelessSetUp()
        >>> setup.setUpTraversal()

        >>> class Stub: pass
        >>> obj = Stub()
        >>> fakePath(obj, '/dir/subdir/name')
        >>> absoluteURL(obj, TestRequest())
        'http://127.0.0.1/dir/subdir/name'

        >>> setup.placelessTearDown()

    """
    folder = FakeRoot()
    bits = [name for name in path.split('/') if name]
    for name in bits[:-1]:
        folder = FakeFolder(folder, name)
    name = bits and bits[-1] or ''
    obj.__parent__ = folder
    obj.__name__ = name



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""Unit tests for schooltool.term.sampledata
"""
import unittest
import doctest

from zope.component import provideAdapter
from zope.interface import Interface
from zope.interface.verify import verifyObject
from zope.app.testing import setup

from schooltool.schoolyear.schoolyear import getSchoolYearContainer
from schooltool.schoolyear.interfaces import ISchoolYearContainer
from schooltool.term.term import getTermContainer
from schooltool.term.interfaces import ITermContainer
from schooltool.testing import setup as stsetup


def setUp(test):
    setup.placefulSetUp()
    provideAdapter(getTermContainer, [Interface], ITermContainer)
    provideAdapter(getSchoolYearContainer)


def tearDown(test):
    setup.placefulTearDown()


def doctest_SampleTerms():
    """A sample data plugin that creates terms

        >>> from schooltool.term.sampledata import SampleTerms
        >>> from schooltool.sampledata.interfaces import ISampleDataPlugin
        >>> plugin = SampleTerms()
        >>> verifyObject(ISampleDataPlugin, plugin)
        True

    This plugin generates a school year:

        >>> app = stsetup.setUpSchoolToolSite()
        >>> plugin.generate(app, 42)
        >>> list(ISchoolYearContainer(app))
        [u'2005-2006']

    The school year contains three terms:

        >>> len(ITermContainer(app))
        3

    These terms are 90 schooldays long:

        >>> fall = ITermContainer(app)['2005-fall']
        >>> schooldays = [day for day in fall if fall.isSchoolday(day)]
        >>> len(schooldays)
        90

        >>> spring = ITermContainer(app)['2006-spring']
        >>> schooldays = [day for day in spring if spring.isSchoolday(day)]
        >>> len(schooldays)
        90

        >>> fall6 = ITermContainer(app)['2006-fall']
        >>> schooldays = [day for day in fall6 if fall6.isSchoolday(day)]
        >>> len(schooldays)
        90

    They span these dates:

        >>> print fall.first, fall.last
        2005-08-22 2005-12-23
        >>> print spring.first, spring.last
        2006-01-26 2006-05-31
        >>> print fall6.first, fall6.last
        2006-08-21 2006-12-22

    """


def test_suite():
    return unittest.TestSuite([
        doctest.DocTestSuite(setUp=setUp, tearDown=tearDown,
                             optionflags=doctest.ELLIPSIS),
        ])


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
SchoolTool Reporlab PDF testing helpers.

"""

import cgi
from cStringIO import StringIO
from lxml import etree

from reportlab import platypus


class StoryXML(object):

    document = None # XML document representing the story

    def __init__(self, story):
        parser = Parser(formatters=_xml_formatters,
                        default=format_classname_xml)
        xml_text = u'<story>\n%s\n</story>' % parser(story)
        self.document = etree.parse(StringIO(xml_text.encode('UTF-8')))
        self._text = etree.tostring(self.document, pretty_print=True)

    def query(self, xpath):
        result = []
        for node in self.document.xpath(xpath):
            if isinstance(node, basestring):
                result.append(node)
            else:
                result.append(etree.tostring(node, pretty_print=True))
        return [s.strip() for s in result]

    def printXML(self, xpath=''):
        if not xpath:
            print self._text
            return
        for entry in self.query(xpath):
            print entry


null_formatter = lambda parser, flowable: u''


class Parser(object):
    def __init__(self, formatters={}, default=null_formatter):
        self.formatters = formatters.copy()
        self.default = default

    def __call__(self, flowable):
        formatter = self.formatters.get(flowable.__class__, self.default)
        return formatter(self, flowable)


def format_flowable_list(parser, flowables):
    parsed = [parser(flowable)
              for flowable in flowables]
    return '\n'.join([text for text in parsed if text])


def format_str_text(parser, flowable):
    return unicode(flowable)


def format_classname_text(parser, flowable):
    return unicode(flowable.__class__.__name__)


def format_str_xml(parser, flowable):
    return cgi.escape(unicode(flowable))


def format_classname_xml(parser, flowable):
    return u'<%s />' % format_classname_text(parser, flowable)


def format_container_xml(parser, flowable):
    tag_name = format_classname_text(parser, flowable)
    content = parser(flowable._content)
    return u'<%s>\n%s\n</%s>' % (tag_name, content, tag_name)


def format_preformatted_xml(parser, flowable):
    tag_name = format_classname_text(parser, flowable)
    return u'<%s bulletText="%s">%s</%s>' % (
        tag_name,
        cgi.escape(flowable.bulletText),
        cgi.escape(u'\n'.join(flowable.lines)),
        tag_name)


def format_table_xml(parser, flowable):
    tag_name = format_classname_text(parser, flowable)
    text = u'<%s>\n' % tag_name
    for row in flowable._cellvalues:
        text += '<tr>\n'
        for cell in row:
            text += '<td>%s</td>\n' % parser(cell)
        text += '</tr>\n'
    text += u'</%s>' % tag_name
    return text


class Format_Attributes_XML(object):
    def __init__(self, attributes=[], content=''):
        self.attribute_names = attributes
        self.content_attribute = content

    def formatAttr(self, parser, flowable, attr_name):
        words = [word for word in attr_name.split('_') if word]
        if words:
            # first word starts with lower case
            words[0] = words[0][:1].lower() + words[0][1:]
        # other words start with upper case
        words[1:] = [word[:1].upper() + word[1:] for word in words[1:]]
        pretty_name = ''.join(words)

        return u'%s="%s"' % (
            pretty_name,
            cgi.escape(str(getattr(flowable, attr_name, None))))

    def formatContents(self, parser, flowable):
        contents = u''
        if self.content_attribute:
            contents = getattr(
                flowable, self.content_attribute, '')
        return unicode(cgi.escape(contents))

    def __call__(self, parser, flowable):
        tag_name = format_classname_text(parser, flowable)
        text = u'<%s' % tag_name
        for attr_name in self.attribute_names:
            text += u' %s' % self.formatAttr(parser, flowable, attr_name)

        contents = self.formatContents(parser, flowable)
        if contents:
            text += u'>%s</%s>' % (contents, tag_name)
        else:
            text += u' />'

        return text


class Format_Paragraph_XML(Format_Attributes_XML):
    def __init__(self, attributes=[]):
        Format_Attributes_XML.__init__(self, attributes=attributes)

    def formatContents(self, parser, flowable):
        return unicode(cgi.escape(flowable.getPlainText()))


class Format_ParaAndImage_XML(Format_Attributes_XML):

    def __init__(self):
        Format_Attributes_XML.__init__(self, ['xpad', 'ypad'])

    def formatContents(self, parser, flowable):
        text = parser([flowable.I, flowable.P])
        return text and '\n%s\n' % text or ''


_xml_formatters = {
    # system
    type(None): null_formatter,
    list: format_flowable_list,

    # plain text
    str: format_str_xml,
    unicode: format_str_xml,

    # paragraph text
    platypus.paragraph.Paragraph: Format_Paragraph_XML(),
    platypus.xpreformatted.XPreformatted: Format_Paragraph_XML(
        attributes=['bulletText']),
    platypus.xpreformatted.PythonPreformatted: Format_Paragraph_XML(
        attributes=['bulletText']),
    platypus.flowables.Preformatted: format_preformatted_xml,

    # graphics
    platypus.flowables.Image:
        Format_Attributes_XML(['filename', '_width', '_height']),
    platypus.flowables.HRFlowable:
        Format_Attributes_XML(
            ['width', 'lineWidth', 'spaceBefore', 'spaceAfter',
             'hAlign', 'vAlign']),

    # containers
    platypus.tables.Table: format_table_xml,
    platypus.tables.LongTable: format_table_xml,
    platypus.flowables.ParagraphAndImage: Format_ParaAndImage_XML(),
    #platypus.flowables.ImageAndFlowables
    #platypus.flowables.PTOContainer, # (Please Turn Over The Page behaviour)

    # spacing
    platypus.flowables.KeepInFrame: format_container_xml,
    platypus.flowables.KeepTogether: format_container_xml,
    platypus.flowables.PageBreak: format_classname_xml,
    platypus.flowables.SlowPageBreak: format_classname_xml,
    platypus.flowables.CondPageBreak: Format_Attributes_XML(['height']),
    platypus.flowables.Spacer: Format_Attributes_XML(
        ['width', 'height']),

    # other
    platypus.flowables.AnchorFlowable: Format_Attributes_XML(['_name']),
    #platypus.tableofcontents.TableOfContents,
    #platypus.tableofcontents.SimpleIndex,

    # omit from output
    platypus.flowables.UseUpSpace: null_formatter,
    platypus.flowables.Flowable: null_formatter,
    platypus.flowables.TraceInfo: null_formatter,
    platypus.flowables.Macro: null_formatter,
    platypus.flowables.CallerMacro: null_formatter,
    platypus.flowables.FailOnWrap: null_formatter,
    platypus.flowables.FailOnDraw: null_formatter,
}




#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2009 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Unit tests for schooltool.generations.evolve37
"""
import unittest
import doctest
import datetime

from persistent.interfaces import IPersistent
from zope.app.testing import setup
from zope.component import provideUtility
from zope.component import provideAdapter
from zope.interface import implements
from zope.intid import IntIds
from zope.intid.interfaces import IIntIds
from zope.keyreference.interfaces import IKeyReference
from zope.site.folder import Folder
from zope.component.hooks import getSite, setSite

from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.generations.tests import ContextStub, StupidKeyReference

from schooltool.schoolyear.schoolyear import SchoolYear
from schooltool.schoolyear.schoolyear import SchoolYearInit
from schooltool.schoolyear.schoolyear import SchoolYearDateRangeAdapter
from schooltool.schoolyear.schoolyear import getSchoolYearContainer
from schooltool.generations.evolve37 import LEVELS_APP_KEY
from schooltool.generations.evolve37 import LevelContainerContainer


class AppStub(Folder):
    implements(ISchoolToolApplication)

    def __init__(self):
        super(AppStub, self).__init__()


def provideAdapters(int_ids):
    provideAdapter(SchoolYearDateRangeAdapter)
    provideAdapter(getSchoolYearContainer)
    provideUtility(int_ids, IIntIds)
    provideAdapter(StupidKeyReference, [IPersistent], IKeyReference)


def setUpYears(app, int_ids):
    SchoolYearInit(app)()
    years = getSchoolYearContainer(app)
    years['1'] = SchoolYear('2010',
                            datetime.date(2010, 01, 01),
                            datetime.date(2010, 12, 30))
    int_ids.register(years['1'])
    years['2'] = SchoolYear('2011',
                            datetime.date(2011, 01, 01),
                            datetime.date(2011, 12, 30))
    int_ids.register(years['2'])
    years._active_id = None
    return years


def doctest_guessMostRecentLevels():
    """Tests for guessing the best level container.

        >>> int_ids = IntIds()
        >>> app = AppStub()
        >>> provideAdapters(int_ids)

        >>> years = setUpYears(app, int_ids)

        >>> from schooltool.generations.evolve37 import guessMostRecentLevels

        >>> print guessMostRecentLevels(app)
        None

        >>> app[LEVELS_APP_KEY] = LevelContainerContainer()

        >>> print guessMostRecentLevels(app)
        None

        >>> app[LEVELS_APP_KEY][unicode(int_ids.getId(years['1']))] = 'levels 1'
        >>> app[LEVELS_APP_KEY][unicode(int_ids.getId(years['2']))] = 'levels 2'

        >>> print guessMostRecentLevels(app)
        levels 2

        >>> years.activateNextSchoolYear(year_id='1')
        >>> print guessMostRecentLevels(app)
        levels 1

        >>> years.activateNextSchoolYear(year_id='2')
        >>> print guessMostRecentLevels(app)
        levels 2

    """


def doctest_evolve37():
    """Test evolution to generation 37.

        >>> context = ContextStub()
        >>> context.root_folder['app'] = app = AppStub()

        >>> manager = setup.createSiteManager(app)

        >>> int_ids = IntIds()
        >>> provideAdapters(int_ids)

        >>> years = setUpYears(app, int_ids)
        >>> years.activateNextSchoolYear(year_id='2')

        >>> app[LEVELS_APP_KEY] = LevelContainerContainer()
        >>> app[LEVELS_APP_KEY][unicode(int_ids.getId(years['1']))] = 'levels 1'
        >>> app[LEVELS_APP_KEY][unicode(int_ids.getId(years['2']))] = 'levels 2'

    Let's evolve now.

        >>> from schooltool.generations.evolve37 import evolve
        >>> evolve(context)

        >>> print app[LEVELS_APP_KEY]
        levels 2

    Site was restored after evolution.

        >>> print getSite()
        None

    """


def doctest_evolve37_no_levels():
    """Test evolution to generation 37.

        >>> context = ContextStub()
        >>> context.root_folder['app'] = app = AppStub()

        >>> manager = setup.createSiteManager(app)

        >>> int_ids = IntIds()
        >>> provideAdapters(int_ids)

        >>> years = setUpYears(app, int_ids)
        >>> years.activateNextSchoolYear(year_id='2')

    Let's evolve now.

        >>> from schooltool.generations.evolve37 import evolve
        >>> evolve(context)

        >>> print app.get(LEVELS_APP_KEY)
        None

    """


def setUp(test):
    setup.placelessSetUp()
    setup.setUpTraversal()
    setSite()

def tearDown(test):
    setSite()
    setup.placelessTearDown()


def test_suite():
    optionflags = (doctest.ELLIPSIS |
                   doctest.NORMALIZE_WHITESPACE |
                   doctest.REPORT_ONLY_FIRST_FAILURE)
    return doctest.DocTestSuite(setUp=setUp, tearDown=tearDown,
                                optionflags=optionflags)


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Pluggable Traverser Implementation
"""

from zope.interface import implements
from zope.component import queryAdapter, queryMultiAdapter
from zope.publisher.interfaces import NotFound

from schooltool.traverser import interfaces


class Traverser(object):
    """A sipmple traverser base."""
    implements(interfaces.IPluggableTraverser)

    def __init__(self, context, request):
        self.context = context
        self.request = request

    def publishTraverse(self, request, name):
        return self.traverse(name)

    def traverse(self, name):
        raise NotFound(self.context, name, self.request)


class PluggableTraverser(object):
    """Generic Pluggable Traverser."""

    implements(interfaces.IPluggableTraverser)

    def __init__(self, context, request):
        self.context = context
        self.request = request

    def publishTraverse(self, request, name):
        # 1. Look for a named traverser plugin.
        named_traverser = queryMultiAdapter((self.context, request),
                                            interfaces.ITraverserPlugin,
                                            name=name)
        if named_traverser is not None:
            try:
                return named_traverser.traverse(name)
            except NotFound:
                pass

        # 2. Named traverser plugin was of no use, let's try a generic one.
        traverser = queryMultiAdapter((self.context, request),
                                      interfaces.ITraverserPlugin)
        if traverser is not None:
            try:
                return traverser.traverse(name)
            except NotFound:
                pass

        # 3. The traversers did not have an answer, so let's see whether it is
        #    a view.
        view = queryMultiAdapter((self.context, request), name=name)
        if view is not None:
            return view

        raise NotFound(self.context, name, request)

    def browserDefault(self, request):
        return self.context, ('index.html', )


class TraverserPlugin(object):
    implements(interfaces.ITraverserPlugin)

    def __init__(self, context, request):
        self.context = context
        self.request = request

    def traverse(self, name):
        raise NotFound(self.context, name, self.request)


class NullTraverserPlugin(TraverserPlugin):
    """Traverse to the context itself (i.e. nowhere)."""

    def traverse(self, name):
        return self.context


class AttributeTraverserPlugin(TraverserPlugin):
    """Traverse to attributes of the context."""

    def traverse(self, name):
        try:
            obj = getattr(self.context, name)
        except AttributeError:
            raise NotFound(self.context, name, self.request)
        else:
            return obj


class AdapterTraverserPlugin(TraverserPlugin):
    """Traverse to an adapter by name."""
    interface = None
    adapterName = ''

    def traverse(self, name):
        adapter = queryAdapter(self.context, self.interface,
                               name=self.adapterName)
        if adapter is None:
            raise NotFound(self.context, name, self.request)

        return adapter


class ContainerTraverserPlugin(TraverserPlugin):
    """A traverser that knows how to look up objects by name in a container."""

    def traverse(self, name):
        subob = self.context.get(name, None)
        if subob is None:
            raise NotFound(self.context, name, self.request)
        return subob



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2012 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
SchoolTool fields.
"""

import zope.schema
from zope.interface import implements
from zope.file.file import File
from zope.schema import Bytes
from zope.schema.interfaces import TooLong

from schooltool.common import SchoolToolMessage as _


_default = object()


class IImage(zope.schema.interfaces.IBytes):

    size = zope.schema.Tuple(
        title=_(u"Image size"),
        value_type = zope.schema.Int(title=_("Size in pixels")),
        required=False
        )

    format = zope.schema.TextLine(
        title=_(u"Image format"),
        required=True)

    max_file_size = zope.schema.Int(
        title=_(u"Max file size in bytes"),
        default=(10 * (10**6)),
        required=True)


class ImageFile(File):
    """A file that is an image."""


class Image(Bytes):

    implements(IImage)
    _type = ImageFile

    size = None
    format = 'PNG'
    max_file_size = 10 * (10**6)

    def __init__(self, size=_default, format=_default,
                 max_file_size=_default, **kw):
        super(Image, self).__init__(**kw)
        if size is not _default:
            self.size = size
        if format is not _default:
            self.format = format
        if max_file_size is not _default:
            self.max_file_size = max_file_size

    def _validate(self, value):
        if self.max_length is not None and value.size > self.max_length:
            raise TooLong(value, self.max_length)




#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
SchoolTool section views
"""

from collections import defaultdict
from urllib import urlencode

import zc.table.table
from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile
from zope.component import adapts, adapter
from zope.component import getMultiAdapter, getAdapter
from zope.component import getUtility
from zope.container.contained import NameChooser
from zope.container.interfaces import INameChooser
from zope.event import notify
from zope.i18n import translate
from zope.i18n.interfaces.locales import ICollator
from zope.interface import implements, Invalid, directlyProvides
from zope.interface import implementer, Interface
from zope.intid.interfaces import IIntIds
from zope.publisher.browser import BrowserView
from zope.publisher.interfaces.browser import IBrowserRequest
from zope.cachedescriptors.property import Lazy
from zope.schema import Choice
from zope.schema import ValidationError
from zope.security.proxy import removeSecurityProxy
from zope.proxy import sameProxiedObjects
from zope.traversing.browser.absoluteurl import absoluteURL
from zope.traversing.browser.interfaces import IAbsoluteURL
from z3c.form import form, subform, field, datamanager, button
from z3c.form.action import ActionErrorOccurred
from z3c.form.interfaces import ActionExecutionError
from z3c.form.interfaces import HIDDEN_MODE
from z3c.form.validator import SimpleFieldValidator
from z3c.form.validator import WidgetValidatorDiscriminators
from zc.table.column import GetterColumn
from zc.table.interfaces import ISortableColumn

from schooltool.app.browser.app import ActiveSchoolYearContentMixin
from schooltool.app.browser.app import BaseEditView
from schooltool.app.browser.app import RelationshipViewBase
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.app.utils import vocabulary_titled
from schooltool.basicperson.browser.person import EditPersonRelationships
from schooltool.basicperson.interfaces import IDemographics
from schooltool.common import SchoolToolMessage as _
from schooltool.common.inlinept import InheritTemplate
from schooltool.common.inlinept import InlineViewPageTemplate
from schooltool.course.interfaces import ICourse, ICourseContainer
from schooltool.course.interfaces import ISection, ISectionContainer
from schooltool.course.section import Section
from schooltool.course.section import copySection
from schooltool.person.interfaces import IPerson
from schooltool.person.interfaces import IPersonFactory
from schooltool.report.browser.report import RequestRemoteReportDialog
from schooltool.resource.browser.resource import EditLocationRelationships
from schooltool.resource.browser.resource import EditEquipmentRelationships
from schooltool.resource.interfaces import ILocation, IEquipment
from schooltool.schoolyear.interfaces import ISchoolYear
from schooltool.schoolyear.interfaces import ISchoolYearContainer
from schooltool.schoolyear.browser.schoolyear import SchoolyearNavBreadcrumbs
from schooltool.securitypolicy.crowds import inCrowd
from schooltool.skin.containers import ContainerView
from schooltool.skin import flourish
from schooltool.skin.flourish.containers import ContainerDeleteView
from schooltool.skin.flourish.form import Dialog
from schooltool.skin.flourish.form import DialogForm
from schooltool.skin.flourish.form import DisplayForm
from schooltool.skin.flourish.form import Form
from schooltool.skin.flourish.page import LinkViewlet
from schooltool.skin.flourish.page import ModalFormLinkViewlet
from schooltool.skin.flourish.page import Page
from schooltool.skin.flourish.page import RefineLinksViewlet
from schooltool.skin.flourish.page import TertiaryNavigationManager
from schooltool import table
from schooltool.term.interfaces import IDateManager
from schooltool.term.interfaces import ITerm
from schooltool.term.term import getPreviousTerm, getNextTerm
from schooltool.term.term import listTerms


class SectionContainerAbsoluteURLAdapter(BrowserView):

    adapts(ISectionContainer, IBrowserRequest)
    implements(IAbsoluteURL)

    def __str__(self):
        container_id = int(self.context.__name__)
        int_ids = getUtility(IIntIds)
        container = int_ids.getObject(container_id)
        url = str(getMultiAdapter((container, self.request), name='absolute_url'))
        return url + '/sections'

    __call__ = __str__


class SectionContainerView(ContainerView):
    """A Course Container view."""

    __used_for__ = ISectionContainer

    @property
    def term(self):
        return ITerm(self.context)

    @property
    def school_year(self):
        return ISchoolYear(self.context)


class SectionCopyingView(SectionContainerView):
    """A view to copy sections from the previous term."""

    __used_for__ = ISectionContainer

    @property
    def container(self):
        term = self.prev_term
        if term is None:
            return None
        return ISectionContainer(term)

    @property
    def prev_term(self):
        return getPreviousTerm(self.term)

    @property
    def select_all_boxes_js(self):
        if self.container is None:
            return ''
        keys = ['"copy.%s"' % section.__name__
                for section in self.batch
                if not section.next]
        return """
            var section_keys = new Array(
                %s
            );

            function selectAll(value) {
                for (key in section_keys) {
                    box = document.getElementById(section_keys[key]);
                    if (box) {
                        box.checked = value;
                    }
                }
            }
        """ % ','.join(keys)

    def update(self):
        if self.container is None:
            return
        if 'COPY' in self.request:
            for name, section in self.container.items():
                key = 'copy.%s' % name
                if key in self.request and not section.next:
                    section = removeSecurityProxy(section)
                    new_section = copySection(section, self.term)
                    section.next = new_section
        SectionContainerView.update(self)


class SectionLinkNextView(SectionContainerView):
    """A view to link a section to one in the next term."""

    __used_for__ = ISectionContainer

    error = u''

    @property
    def container(self):
        term = self.target_term
        if term is None:
            return None
        return ISectionContainer(term)

    @property
    def target_term(self):
        return getNextTerm(self.term)

    def link(self, target_section):
        section = removeSecurityProxy(self.context)
        target_section = removeSecurityProxy(target_section)
        section.next = target_section

    def update(self):
        if 'CANCEL' in self.request:
            self.request.response.redirect(self.nextURL())
        if self.container is None:
            return
        if 'LINK' in self.request:
            section_name = self.request.get('LINK_SECTION')
            section = self.container.get(section_name)
            if section_name is None or section is None:
                self.error = translate(_('No section selected.'),
                                       context=self.request)
            else:
                self.link(section)
                self.request.response.redirect(self.nextURL())
        SectionContainerView.update(self)

    def nextURL(self):
        return '%s/@@edit.html' % absoluteURL(self.context, self.request)


class SectionLinkPreviousView(SectionLinkNextView):
    """A view to link a section to one in the previous term."""

    @property
    def target_term(self):
        return getPreviousTerm(self.term)

    def link(self, target_section):
        section = removeSecurityProxy(self.context)
        target_section = removeSecurityProxy(target_section)
        section.previous = target_section


class SectionView(BrowserView):
    """A view for courses providing a list of sections."""

    __used_for__ = ISection

    @property
    def term(self):
        return ITerm(self.context)

    @property
    def school_year(self):
        return ISchoolYear(self.context)

    @property
    def linked_terms(self):
        sections = []
        current = self.context
        while current.previous:
            sections.append(current.previous)
            current = current.previous
        sections.reverse()
        current = self.context
        while current.next:
            sections.append(current.next)
            current = current.next
        for section in sections:
            yield {
                'section': section,
                'term': ITerm(section),
                }

    def renderPersonTable(self):
        persons = ISchoolToolApplication(None)['persons']
        formatter = getMultiAdapter((persons, self.request),
                                    table.interfaces.ITableFormatter)
        formatter.setUp(table_formatter=zc.table.table.StandaloneFullFormatter,
                        items=[removeSecurityProxy(person)
                               for person in self.context.members],
                        batch_size=0)
        return formatter.render()


class SectionNameChooser(NameChooser):

    implements(INameChooser)

    def chooseName(self, name, obj):
        """See INameChooser."""

        i = 1
        n = "1"
        while n in self.context:
            i += 1
            n = unicode(i)
        # Make sure the name is valid
        self.checkName(n, obj)
        return n


class SectionAddView(form.AddForm):
    """A view for adding Sections."""

    default_term = None
    default_course = None
    subforms = None
    _created_obj = None

    template = ViewPageTemplateFile('templates/section_add.pt')

    # Note that we also omit the title field, it will be auto-generated
    # See concerns raised in https://bugs.launchpad.net/schooltool/+bug/389283
    fields = field.Fields(ISection).omit(
        '__name__', 'title',
        'label', 'instructors', 'members', 'courses', 'size',
        'previous', 'next', 'linked_sections')

    def update(self):
        super(SectionAddView, self).update()
        self._finishedAdd = False

        self.course_subform = NewSectionCoursesSubform(
            self.context, self.request, self,
            default_course=self.default_course)
        self.term_subform = NewSectionTermsSubform(
            self.context, self.request, self,
            default_term=self.default_term)

        self.course_subform.update()
        self.term_subform.update()

        if (self.course_subform.errors or
            self.term_subform.errors or
            self.widgets.errors):
            self.status = self.formErrorsMessage
        elif self._created_obj:
            self.add(self._created_obj)
            self._finishedAdd = True

    def create(self, data):
        section = Section()
        form.applyChanges(self, section, data)
        return section

    def add(self, section):
        """Add `contact` to the container.

        Uses the username of `contact` as the object ID (__name__).
        """

        course = self.course_subform.course
        terms = self.term_subform.terms
        if course is None or not terms:
            return None

        # add the section to the first term
        sections = ISectionContainer(terms[0])
        name = INameChooser(sections).chooseName('', section)
        sections[name] = section
        section.courses.add(removeSecurityProxy(course))

        # overwrite section title.
        section.title = "%s (%s)" % (course.title, section.__name__)

        # copy and link section in other selected terms
        for term in terms[1:]:
            new_section = copySection(section, term)
            new_section.previous = section
            section = new_section
        self._finishedAdd = False

    def nextURL(self):
        return absoluteURL(self.context, self.request)

    @button.buttonAndHandler(_('Add'), name='add')
    def handleAdd(self, action):
        data, errors = self.extractData()
        if errors:
            self.status = self.formErrorsMessage
            return
        # Note we skip immediate adding of the object,
        # because we want to check subforms for errors first.
        self._created_obj = self.create(data)

    @button.buttonAndHandler(_("Cancel"), name='cancel')
    def handle_cancel_action(self, action):
        url = absoluteURL(self.context, self.request)
        self.request.response.redirect(self.nextURL())

    def updateActions(self):
        super(SectionAddView, self).updateActions()
        self.actions['add'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')


class NewSectionTermsSubform(subform.EditSubForm):
    template = ViewPageTemplateFile('templates/basic_subform.pt')
    prefix = 'terms'

    errors = None
    span = None

    def __init__(self, *args, **kw):
        default_term = kw.pop('default_term', None)
        super(NewSectionTermsSubform, self).__init__(*args, **kw)

        if default_term is None:
            selected_year = removeSecurityProxy(ISchoolYear(self.context))
            active_term = getUtility(IDateManager).current_term
            if (active_term is not None and
                ISchoolYear(active_term) is selected_year):
                default_term = active_term

        self.setUpFields(default_term)

    def setUpFields(self, default_term):
        terms = listTerms(self.context)
        self.vocabulary=vocabulary_titled(terms)
        self.span = defaultdict(lambda:default_term)
        self._addTermChoice('starts', _("Starts in term"))
        self._addTermChoice('ends', _("Ends in term"))

    def _addTermChoice(self, name, title):
        schema_field = Choice(
            __name__=name, title=title,
            required=True,
            vocabulary=self.vocabulary,
            default=self.span[name])

        self.fields += field.Fields(schema_field)

    def getContent(self):
        return self.span

    @property
    def terms(self):
        if self.errors:
            return []
        terms = listTerms(self.context)
        if (self.span['starts'] is None or
            self.span['ends'] is None):
            return []
        return [term for term in terms
                if (term.first >= self.span['starts'].first and
                    term.last <= self.span['ends'].last)]

    @button.handler(SectionAddView.buttons['add'])
    def handleAdd(self, action):
        data, self.errors = self.widgets.extract()
        if self.errors:
            return
        changed = form.applyChanges(self, self.getContent(), data)
        starts = self.span['starts']
        ends = self.span['ends']
        if starts.first > ends.first:
            # XXX: this is a workaround for a bug in z3c.form: subforms do
            #      not handle action execution errors properly.
            #      The bug is fixed in z3c.form 2.0, as far as I know.
            widget_name = '%s.widgets.ends:list' % self.prefix
            error = ActionExecutionError(Invalid(
                _('Starting term ($starts_in) is later than ending term ($ends_in)',
                  mapping={'starts_in': starts.title,
                           'ends_in': ends.title})))
            notify(ActionErrorOccurred(action, error))


# XXX: TODO: Add "--no value--" to the course selector dropdown in
#      NewSectionCoursesSubform, once we depend on z3c.form 2.0.


class NewSectionCoursesSubform(subform.EditSubForm):
    template = ViewPageTemplateFile('templates/basic_subform.pt')
    prefix = 'courses'

    errors = None
    values = None

    def __init__(self, *args, **kw):
        default_course = kw.pop('default_course', None)
        super(NewSectionCoursesSubform, self).__init__(*args, **kw)
        courses = ICourseContainer(self.context)
        self.vocabulary=vocabulary_titled(courses.values())
        self.values = {'course': default_course}
        schema_field = Choice(
            __name__='course', title=_('Course'),
            required=True, vocabulary=self.vocabulary)
        self.fields += field.Fields(schema_field)
        datamanager.DictionaryField(self.values, schema_field)

    def getContent(self):
        return self.values

    @property
    def course(self):
        return self.values.get('course')

    @button.handler(SectionAddView.buttons['add'])
    def handleAdd(self, action):
        data, self.errors = self.widgets.extract()
        if self.errors:
            return
        changed = form.applyChanges(self, self.getContent(), data)


class AddSectionForTerm(SectionAddView):
    """A view for adding Sections for a term."""

    form.extends(SectionAddView)

    def __init__(self, *args, **kw):
        super(AddSectionForTerm, self).__init__(*args, **kw)
        self.default_term = ITerm(self.context)


class AddSectionForCourse(SectionAddView):
    """A view for adding Sections for a course."""

    form.extends(SectionAddView)

    def __init__(self, *args, **kw):
        super(AddSectionForCourse, self).__init__(*args, **kw)
        self.default_course = ICourse(self.context)


class SectionEditView(BaseEditView):
    """View for editing Sections."""

    __used_for__ = ISection

    @property
    def next_term(self):
        return getNextTerm(ITerm(self.context))

    @property
    def previous_term(self):
        return getPreviousTerm(ITerm(self.context))

    def update(self):
        if 'UNLINK_NEXT' in self.request:
            section = removeSecurityProxy(self.context)
            section.next = None
            return ''
        elif 'UNLINK_PREVIOUS' in self.request:
            section = removeSecurityProxy(self.context)
            section.previous = None
            return ''
        return BaseEditView.update(self)


class RelationshipEditConfView(RelationshipViewBase):
    """A relationship editing view that displays conflicts."""

    __call__ = ViewPageTemplateFile('templates/edit_relationships.pt')

    @property
    def term(self):
        return ITerm(self.context)

    @property
    def school_year(self):
        return ISchoolYear(self.context)


class SectionInstructorView(RelationshipEditConfView):
    """View for adding instructors to a Section."""

    __used_for__ = ISection

    title = _("Instructors")
    current_title = _("Current Instructors")
    available_title = _("Available Instructors")

    def getCollection(self):
        return self.context.instructors

    def getAvailableItemsContainer(self):
        return ISchoolToolApplication(None)['persons']


class SectionLearnerView(RelationshipEditConfView):
    """View for adding learners to a Section.  """

    __used_for__ = ISection

    title = _("Students")
    current_title = _("Current Students")
    available_title = _("Available Students")

    def getCollection(self):
        return self.context.members

    def getSelectedItems(self):
        """Return a list of selected members."""
        return filter(IPerson.providedBy, self.getCollection())

    def getAvailableItemsContainer(self):
        return ISchoolToolApplication(None)['persons']


class SectionLinkageView(BrowserView):
    """A view for seeing all terms of a section and linking it to others."""

    @property
    def year(self):
        return ISchoolYear(self.context)

    @property
    def linked(self):
        return len(self.context.linked_sections) > 1

    @property
    def columns(self):
        linked_sections = dict([(ITerm(section), section)
                                 for section in self.context.linked_sections])
        columns = []
        for term in sorted(self.year.values(), key=lambda t: t.first):
            columns.append({
                'term': term,
                'section': linked_sections.get(term),
                })
        return columns


class ExtendTermView(BrowserView):
    """A view for extending a section to a target term."""

    template = ViewPageTemplateFile('templates/extend_term.pt')

    @property
    def term(self):
        return ITerm(self.context)

    @property
    def extend_term(self):
        return ISchoolYear(self.context).get(self.request['term'])

    def __call__(self):
        section = removeSecurityProxy(self.context)
        year = ISchoolYear(section)
        linked_names = [ITerm(s).__name__ for s in section.linked_sections]
        key = self.request['term']

        if key not in year or key in linked_names or 'CANCEL' in self.request:
            self.request.response.redirect(self.nextURL())

        elif 'EXTEND' in self.request:
            this_term = ITerm(self.context)
            extend_term = year[key]
            if extend_term.first < this_term.first:
                current = section.linked_sections[0]
                target_term = getPreviousTerm(ITerm(current))
            else:
                current = section.linked_sections[-1]
                target_term = getNextTerm(ITerm(current))
            while ITerm(current).first != extend_term.first:
                new_section = copySection(current, target_term)
                if extend_term.first < this_term.first:
                    new_section.next = current
                    target_term = getPreviousTerm(target_term)
                else:
                    new_section.previous = current
                    target_term = getNextTerm(target_term)
                current = new_section
            self.request.response.redirect(self.nextURL())

        else:
            return self.template()

    def nextURL(self):
        return absoluteURL(self.context, self.request) + '/section_linkage.html'


class LinkExistingView(BrowserView):
    """A view for finding a target section in anoother term for linking."""

    template = ViewPageTemplateFile('templates/link_existing.pt')

    @property
    def term(self):
        return ITerm(self.context)

    @property
    def link_term(self):
        term = ISchoolYear(self.context).get(self.request.get('term'))
        if (term is not None and
            term not in [ITerm(s) for s in self.context.linked_sections]):
            return term

    @property
    def sections(self):
        section = removeSecurityProxy(self.context)
        term = self.link_term

        sections = []
        if term is None:
            return sections

        courses = list(section.courses)
        for target in ISectionContainer(term).values():
            if target in section.linked_sections:
                continue
            if not self.filterSection(target):
                continue
            if courses == list(target.courses):
                sections.append(target)
        return sections

    def filterSection(self, section):
        teacher_filter = self.request.get('teacher', '').lower()
        if not teacher_filter:
            return True
        for teacher in section.instructors:
            name = teacher.title
            if (teacher_filter in name.lower() or
                teacher_filter in teacher.username.lower()):
                return True
        return False

    @property
    def selected_section(self):
        term = self.link_term
        selection = self.request.get('LINK_SECTION')
        if term is not None and selection is not None:
            section = ISectionContainer(term).get(selection[8:])
            if section not in self.context.linked_sections:
                return section

    @property
    def error(self):
        return 'LINK' in self.request and self.selected_section is None

    def __call__(self):
        section = removeSecurityProxy(self.context)
        term = self.term
        link_term = self.link_term

        if link_term is None or 'CANCEL' in self.request:
            self.request.response.redirect(self.nextURL())

        elif 'LINK' in self.request:
            target_section = self.selected_section
            if target_section is None:
                return self.template()

            if link_term.first < term.first:
                current = section.linked_sections[0]
                target_term = getPreviousTerm(ITerm(current))
            else:
                current = section.linked_sections[-1]
                target_term = getNextTerm(ITerm(current))

            while target_term.first != link_term.first:
                new_section = copySection(current, target_term)
                if link_term.first < term.first:
                    new_section.next = current
                    target_term = getPreviousTerm(target_term)
                else:
                    new_section.previous = current
                    target_term = getNextTerm(target_term)
                current = new_section

            if link_term.first < term.first:
                target_section.next = current
            else:
                target_section.previous = current

            self.request.response.redirect(self.nextURL())

        else:
            return self.template()

    def nextURL(self):
        return absoluteURL(self.context, self.request) + '/section_linkage.html'


class UnlinkSectionView(BrowserView):
    """A view for unlinking a section."""

    template = ViewPageTemplateFile('templates/unlink_section.pt')

    @property
    def term(self):
        return ITerm(self.context)

    @property
    def sections(self):
        sections = []
        for section in [self.context.previous, self.context.next]:
            if section:
                sections.append({
                    'section': section,
                    'term': ITerm(section),
                    })
        return sections

    def __call__(self):
        section = removeSecurityProxy(self.context)

        if not self.sections or 'CANCEL' in self.request:
            self.request.response.redirect(self.nextURL())

        elif 'UNLINK' in self.request:
            if section.previous:
                section.previous = None
            if section.next:
                section.next = None
            self.request.response.redirect(self.nextURL())

        else:
            return self.template()

    def nextURL(self):
        return absoluteURL(self.context, self.request) + '/section_linkage.html'


class SectionsTertiaryNavigationManager(TertiaryNavigationManager,
                                        ActiveSchoolYearContentMixin):

    template = InlineViewPageTemplate("""
        <ul tal:attributes="class view/list_class">
          <li tal:repeat="item view/items"
              tal:attributes="class item/class"
              tal:content="structure item/viewlet">
          </li>
        </ul>
    """)

    @property
    def items(self):
        result = []
        active = self.schoolyear
        schoolyears = active.__parent__ if active is not None else {}
        for schoolyear in schoolyears.values():
            url = '%s/%s?schoolyear_id=%s' % (
                absoluteURL(self.context, self.request),
                'sections',
                schoolyear.__name__)
            result.append({
                    'class': schoolyear.first == active.first and 'active' or None,
                    'viewlet': u'<a href="%s">%s</a>' % (url, schoolyear.title),
                    })
        return result


class SectionsAddLinks(RefineLinksViewlet):
    """Manager for Add links in SectionsView"""


class SectionImportLinks(RefineLinksViewlet):
    """Section import links viewlet."""


class SectionLinks(RefineLinksViewlet):
    """Manager for public links in SectionView"""

    @property
    def title(self):
        return self.context.title


class SectionAddLinks(RefineLinksViewlet):
    """Manager for Add links in SectionView"""


class SectionActionsLinks(RefineLinksViewlet):
    """Manager for Action links in SectionView"""

    body_template = InlineViewPageTemplate("""
        <ul tal:attributes="class view/list_class">
          <li tal:repeat="item view/renderable_items"
              tal:attributes="class item/class"
              tal:content="structure item/viewlet">
          </li>
        </ul>
    """)

    # We don't want this manager rendered at all
    # if there are no renderable viewlets
    @property
    def renderable_items(self):
        result = []
        for item in self.items:
            render_result = item['viewlet']()
            if render_result and render_result.strip():
                result.append({
                        'class': item['class'],
                        'viewlet': render_result,
                        })
        return result

    def render(self):
        if self.renderable_items:
            return super(SectionActionsLinks, self).render()


class SectionAddLinkViewlet(LinkViewlet, ActiveSchoolYearContentMixin):

    @property
    def enabled(self):
        if (not flourish.canEdit(self.context) and
            not inCrowd(self.request.principal, 'clerks', context=self.context)):
            return False
        return super(SectionAddLinkViewlet, self).enabled

    @property
    def url(self):
        return '%s/%s' % (absoluteURL(self.schoolyear, self.request),
                          'addSection.html')


class SectionAddLinkFromSectionViewlet(SectionAddLinkViewlet):

    @property
    def enabled(self):
        container = self.context.__parent__
        if not flourish.canEdit(container):
            return False
        return super(SectionAddLinkViewlet, self).enabled

    @property
    def schoolyear(self):
        return ISchoolYear(self.context)

    @property
    def url(self):
        return '%s/%s?camefrom=%s' % (
            absoluteURL(self.schoolyear, self.request),
            'addSection.html',
            absoluteURL(self.context, self.request))


class SectionDeleteLink(ModalFormLinkViewlet):

    @property
    def enabled(self):
        if not flourish.canDelete(self.context):
            return False
        return super(SectionDeleteLink, self).enabled

    @property
    def dialog_title(self):
        title = _(u'Delete ${section}',
                  mapping={'section': self.context.title})
        return translate(title, context=self.request)


class FlourishSectionFilterWidget(table.table.FilterWidget):

    template = ViewPageTemplateFile('templates/f_section_filter.pt')


class FlourishSectionTableFormatter(table.table.SchoolToolTableFormatter):

    def columns(self):
        title = table.column.LocaleAwareGetterColumn(
            name='title',
            title=_(u'Title'),
            getter=lambda i, f: i.title,
            subsort=True)
        directlyProvides(title, ISortableColumn)
        return [title]

    def filter(self, items):
        filter_widget = FlourishSectionFilterWidget(self.context, self.request)
        self.filter_widget = filter_widget
        return filter_widget.filter(items)


def get_courses_titles(section, formatter):
    return ', '.join([course.title for course in section.courses])


def get_section_instructors(section, formatter):
    return ', '.join([person.title for person in section.instructors])


def section_instructors_formatter(value, section, formatter):
    return '<br />'.join([person.title for person in section.instructors])


class FlourishSectionsView(flourish.page.Page,
                           ActiveSchoolYearContentMixin):

    container_class = 'container widecontainer'
    content_template = InlineViewPageTemplate('''
      <div tal:content="structure context/schooltool:content/ajax/view/schoolyear/sections_table" />
    ''')

    @property
    def title(self):
        schoolyear = self.schoolyear
        return _('Sections for ${schoolyear}',
                 mapping={'schoolyear': schoolyear.title})


class SectionsTableBase(table.ajax.Table):

    def columns(self):
        default = table.ajax.Table.columns(self)
        term = table.column.LocaleAwareGetterColumn(
            name='term',
            title=_('Term'),
            getter=lambda i, f: ITerm(i).title,
            subsort=True)
        courses = table.column.LocaleAwareGetterColumn(
            name='courses',
            title=_('Courses'),
            getter=get_courses_titles,
            subsort=True)
        size = GetterColumn(
            name='size',
            title=_('Students'),
            getter=lambda i, f: i.size,
            subsort=True)
        instructors = table.column.LocaleAwareGetterColumn(
            name='instructors',
            title=_('Teachers'),
            getter=get_section_instructors,
            cell_formatter=section_instructors_formatter)
        directlyProvides(term, ISortableColumn)
        directlyProvides(courses, ISortableColumn)
        directlyProvides(instructors, ISortableColumn)
        return default + [term, courses, instructors, size]

    def sortOn(self):
        return (('term', True), ('courses', False), ('title', False))


class SectionsTable(SectionsTableBase):

    pass


class SectionListTable(SectionsTableBase):

    def columns(self):
        default = super(SectionListTable, self).columns()
        title, term, courses, instructors, size = default
        return [title, term, instructors]

    def sortOn(self):
        return (('term', True), ('title', False))

    @Lazy
    def source(self):
        sections = {}
        schoolyear = ISchoolYear(self.context)
        for term in schoolyear.values():
            term_section_container = ISectionContainer(term)
            for section in term_section_container.values():
                name = '%s.%s.%s' % (
                    schoolyear.__name__, term.__name__, section.__name__
                    )
                sections[name] = section
        return sections

    def items(self):
        if flourish.canView(self.context):
            return removeSecurityProxy(self.context).sections
        return ()


class SchoolYearSectionsTable(SectionsTable):

    @property
    def schoolyear(self):
        return ISchoolYear(self.context)

    @Lazy
    def source(self):
        sections = {}
        schoolyear = self.schoolyear
        for term in schoolyear.values():
            term_section_container = ISectionContainer(term)
            for section in term_section_container.values():
                name = '%s.%s.%s' % (
                    schoolyear.__name__, term.__name__, section.__name__
                    )
                sections[name] = section
        return sections


class SectionsTableFilter(table.ajax.TableFilter,
                          FlourishSectionFilterWidget,
                          ActiveSchoolYearContentMixin):

    multiple_terms = True
    template = ViewPageTemplateFile('templates/f_section_table_filter.pt')
    title = _("Section title")

    @property
    def search_id(self):
        return self.manager.html_id+'-search'

    @property
    def search_title_id(self):
        return self.manager.html_id+"-title"

    @property
    def search_course_id(self):
        return self.manager.html_id+"-course"

    @property
    def search_term_ids(self):
        return self.manager.html_id+"-terms"

    def termContainer(self):
        return self.schoolyear

    def courseContainer(self):
        return ICourseContainer(self.schoolyear)

    def terms(self):
        result = []
        container = self.termContainer()
        items = sorted(container.items(),
                       key=lambda (tid, term):term.first)
        for id, term in items:
            checked = not self.manager.fromPublication
            if self.search_term_ids in self.request:
                term_ids = self.request[self.search_term_ids]
                if not isinstance(term_ids, list):
                    term_ids = [term_ids]
                checked = id in term_ids
            result.append({'id': id,
                           'title': term.title,
                           'checked': checked,
                           'obj': term})
        return result

    def courses(self):
        result = []
        container = self.courseContainer()
        collator = ICollator(self.request.locale)
        items = sorted(container.items(),
                       cmp=collator.cmp,
                       key=lambda (cid, c): c.title)
        for id, course in items:
            result.append({'id': id,
                           'title': course.title})
        return result

    def filter(self, items):
        if len(self.termContainer()) < 2:
            self.multiple_terms = False
        if self.ignoreRequest:
            return items
        if self.search_term_ids in self.request:
            term_ids = self.request[self.search_term_ids]
            if not isinstance(term_ids, list):
                term_ids = [term_ids]
            terms = []
            for term_id in term_ids:
                term = self.termContainer().get(term_id)
                if term is not None:
                    terms.append(term)
            if terms:
                items = [item for item in items
                         if ITerm(item) in terms]
        elif self.multiple_terms:
            return []
        if self.search_course_id in self.request:
            course_id = self.request[self.search_course_id]
            course = self.courseContainer().get(course_id)
            if course:
                items = [item for item in items
                         if item in course.sections]
        if self.search_title_id in self.request:
            searchstr = self.request[self.search_title_id].lower()
            items = [item for item in items
                     if searchstr in item.title.lower()]
        return items


class SectionListTableFilter(SectionsTableFilter):

    template = ViewPageTemplateFile('templates/f_section_list_table_filter.pt')

    @Lazy
    def schoolyear(self):
        return ISchoolYear(self.context)

    def getSectionCount(self, term):
        return len([section for section in ISectionContainer(term).values()
                    if section in self.context.sections])


class SectionsTableSchoolYear(flourish.viewlet.Viewlet):

    template = InlineViewPageTemplate('''
      <input type="hidden" name="schoolyear_id"
             tal:define="schoolyear_id view/view/schoolyear/__name__|nothing"
             tal:condition="schoolyear_id"
             tal:attributes="value schoolyear_id" />
    ''')


class FlourishSectionContainerDeleteView(ContainerDeleteView):

    def nextURL(self):
        if 'CONFIRM' in self.request:
            schoolyear = ISchoolYear(self.context)
            params = {'schoolyear_id': schoolyear.__name__.encode('utf-8')}
            url = '%s/sections?%s' % (
                absoluteURL(ISchoolToolApplication(None), self.request),
                urlencode(params))
            return url
        return ContainerDeleteView.nextURL(self)


class FlourishSectionView(DisplayForm, ActiveSchoolYearContentMixin):

    template = InheritTemplate(Page.template)
    content_template = ViewPageTemplateFile('templates/f_section_view.pt')
    fields = field.Fields(ISection)
    fields = fields.select('__name__', 'title', 'description')

    @property
    def courses(self):
        return list(self.context.courses)

    @property
    def show_extended_roster(self):
        if flourish.canEdit(self.context):
            return True
        persons = ISchoolToolApplication(None)['persons']
        return flourish.canView(persons)

    @property
    def linked_terms(self):
        sections = []
        current = self.context
        while current.previous:
            sections.append(current.previous)
            current = current.previous
        sections.reverse()
        current = self.context
        sections.append(current)
        while current.next:
            sections.append(current.next)
            current = current.next
        return [
            {'term': ITerm(section),
             'section': section,
             'current': section is self.context,
             }
            for section in sections]

    @property
    def schoolyear(self):
        return ISchoolYear(self.context)

    @property
    def title(self):
        return _('Sections for ${schoolyear}',
                 mapping={'schoolyear': self.schoolyear.title})

    @property
    def subtitle(self):
        return self.context.title

    def done_link(self):
        app = ISchoolToolApplication(None)
        url = self.url_with_schoolyear_id(app, view_name='sections')
        return url

    def updateWidgets(self):
        super(FlourishSectionView, self).updateWidgets()
        for widget in self.widgets.values():
            if not widget.value:
                widget.mode = HIDDEN_MODE

    def has_instructors(self):
        return bool(list(self.context.instructors))

    def has_learners(self):
        return bool(list(self.context.members))

    def instructor_titles(self):
        instructors = []
        if flourish.canView(self.context):
            instructors = list(removeSecurityProxy(self.context).instructors)
        titles = [getMultiAdapter((instructor, self.request), name='title')()
                  for instructor in instructors]
        return titles

    def student_titles(self):
        students = []
        if flourish.canView(self.context):
            students = list(removeSecurityProxy(self.context).members)
        titles = [getMultiAdapter((student, self.request), name='title')()
                  for student in students]
        return titles

    def has_locations(self):
        return bool([r for r in self.context.resources
                     if ILocation(r, None) is not None])

    def has_equipment(self):
        return bool([r for r in self.context.resources
                     if IEquipment(r, None) is not None])


class ISectionAddTitleHint(Interface):
    """Section add optional title hint text."""


@adapter(ISchoolYear)
@implementer(ISectionAddTitleHint)
def getSectionAddTitleHint(context):
    return _(u"If no title is specified, one will be created based on the "
              "course title.")


class FlourishSectionAddView(Form, SectionAddView, ActiveSchoolYearContentMixin):

    template = InheritTemplate(Page.template)
    label = None
    legend = _('Section Information')

    fields = field.Fields(ISection).select('title', 'description')

    @property
    def schoolyear(self):
        return self.context

    @property
    def title(self):
        return _('Sections for ${schoolyear}',
                 mapping={'schoolyear': self.schoolyear.title})

    def updateWidgets(self):
        super(FlourishSectionAddView, self).updateWidgets()
        self.widgets['title'].required = False
        self.widgets['title'].field.required = False
        self.widgets['title'].field.description = getAdapter(self.context,
            ISectionAddTitleHint)

    def update(self):
        super(SectionAddView, self).update()
        self._finishedAdd = False

        self.course_subform = FlourishNewSectionCoursesSubform(
            self.context, self.request, self,
            default_course=self.default_course)
        self.term_subform = FlourishNewSectionTermsSubform(
            self.context, self.request, self,
            default_term=self.default_term)
        self.location_subform = FlourishNewSectionLocationSubform(
            self.context, self.request, self)

        self.course_subform.update()
        self.term_subform.update()
        self.location_subform.update()

        if (self.course_subform.errors or
            self.term_subform.errors or
            self.location_subform.errors or
            self.widgets.errors):
            self.errors = self.course_subform.errors  + \
                          self.term_subform.errors + \
                          self.location_subform.errors + \
                          self.widgets.errors
            self.status = self.formErrorsMessage
        elif self._created_obj:
            self.add(self._created_obj)
            self._finishedAdd = True

    def add(self, section):
        course = self.course_subform.course
        terms = self.term_subform.terms
        location = self.location_subform.location
        if course is None or not terms:
            return None

        # add the section to the first term
        sections = ISectionContainer(terms[0])
        name = INameChooser(sections).chooseName('', section)
        sections[name] = section
        self._section = section
        section.courses.add(removeSecurityProxy(course))

        # if user provides no title, set it to default
        if not section.title:
            section.title = u"%s (%s)" % (course.title, section.__name__)

        # copy and link section in other selected terms
        for term in terms[1:]:
            new_section = copySection(section, term)
            new_section.previous = section
            section = new_section

        # if the user provides a location, add it
        if location is not None:
            for section in self._section.linked_sections:
                section.resources.add(removeSecurityProxy(location))

        self._finishedAdd = False

    def nextURL(self):
        return absoluteURL(self._section, self.request)

    @button.buttonAndHandler(_('Submit'), name='add')
    def handleAdd(self, action):
        data, errors = self.extractData()
        if errors:
            self.status = self.formErrorsMessage
            return
        # Note we skip immediate adding of the object,
        # because we want to check subforms for errors first.
        self._created_obj = self.create(data)

    @button.buttonAndHandler(_("Cancel"), name='cancel')
    def handle_cancel_action(self, action):
        if 'camefrom' in self.request:
            url = self.request['camefrom']
            self.request.response.redirect(url)
            return
        app = ISchoolToolApplication(None)
        url = self.url_with_schoolyear_id(app, view_name='sections')
        self.request.response.redirect(url)


class FlourishNewSectionTermsSubform(NewSectionTermsSubform):

    @button.handler(FlourishSectionAddView.buttons['add'])
    def handleAdd(self, action):
        data, self.errors = self.widgets.extract()
        if self.errors:
            return
        changed = form.applyChanges(self, self.getContent(), data)


class FlourishSectionTermError(ValidationError):
    __doc__ = _('Starting term is later than ending term')


class SectionTermsValidator(SimpleFieldValidator):

    def validate(self, value):
        # XXX: hack to display the term error next to the widget!
        if self.widget.__name__ == 'starts':
            starts_term = value
            super(SectionTermsValidator, self).validate(starts_term)
            ends_widget = self.view.widgets['ends']
            ends_value = self.request.get(ends_widget.name)[0]
            ends_term = ends_widget.terms.getTermByToken(ends_value).value
            if starts_term.first > ends_term.first:
                raise FlourishSectionTermError()


WidgetValidatorDiscriminators(SectionTermsValidator,
                              view=FlourishNewSectionTermsSubform)


class FlourishNewSectionCoursesSubform(NewSectionCoursesSubform):

    def __init__(self, *args, **kw):
        default_course = kw.pop('default_course', None)
        subform.EditSubForm.__init__(self, *args, **kw)
        courses = ICourseContainer(self.context)
        collator = ICollator(self.request.locale)
        items = sorted(courses.values(),
                       cmp=collator.cmp,
                       key=lambda course:course.title)
        self.vocabulary=vocabulary_titled(items)
        self.values = {'course': default_course}
        schema_field = Choice(
            __name__='course', title=_('Course'),
            required=True, vocabulary=self.vocabulary)
        self.fields += field.Fields(schema_field)
        datamanager.DictionaryField(self.values, schema_field)


    def updateWidgets(self):
        super(FlourishNewSectionCoursesSubform, self).updateWidgets()
        self.widgets['course'].prompt = True
        self.widgets['course'].promptMessage = _('Select a course')

    @button.handler(FlourishSectionAddView.buttons['add'])
    def handleAdd(self, action):
        data, self.errors = self.widgets.extract()
        if self.errors:
            return
        changed = form.applyChanges(self, self.getContent(), data)


class ISectionAddLocationHint(Interface):
    """Section add optional location hint text."""


@adapter(ISchoolYear)
@implementer(ISectionAddLocationHint)
def getSectionAddLocationHint(context):
    return _(u"Additional locations can be added via the section view.")


class FlourishNewSectionLocationSubform(subform.EditSubForm):

    template = ViewPageTemplateFile('templates/section_add_location_subform.pt')
    prefix = 'location'
    errors = None

    def __init__(self, *args, **kw):
        super(FlourishNewSectionLocationSubform, self).__init__(*args, **kw)
        self.values = {'location': None}

        # add location field with titled lacation resources vocabulary
        app = ISchoolToolApplication(None)
        resources = [r for r in app['resources'].values()
                     if ILocation(r, None) is not None]
        schema_field = Choice(
            __name__='location', title= _("Location"),
            required=False, vocabulary=vocabulary_titled(resources))
        self.fields += field.Fields(schema_field)
        datamanager.DictionaryField(self.values, schema_field)

    def updateWidgets(self):
        super(FlourishNewSectionLocationSubform, self).updateWidgets()
        self.widgets['location'].field.description = getAdapter(self.context,
            ISectionAddLocationHint)

    def getContent(self):
        return self.values

    @property
    def location(self):
        return self.values.get('location')

    @button.handler(FlourishSectionAddView.buttons['add'])
    def handleAdd(self, action):
        data, self.errors = self.widgets.extract()
        if self.errors:
            return
        changed = form.applyChanges(self, self.getContent(), data)


class FlourishSectionEditView(Form, form.EditForm):

    template = InheritTemplate(Page.template)
    label = None
    legend = _('Section Information')
    fields = field.Fields(ISection).select('title', 'description')

    @property
    def title(self):
        return self.context.title

    def update(self):
        return form.EditForm.update(self)

    def updateActions(self):
        super(FlourishSectionEditView, self).updateActions()
        self.actions['apply'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')

    @button.buttonAndHandler(_('Submit'), name='apply')
    def handleApply(self, action):
        super(FlourishSectionEditView, self).handleApply.func(self, action)
        # XXX: hacky sucessful submit check
        if (self.status == self.successMessage or
            self.status == self.noChangesMessage):
            url = absoluteURL(self.context, self.request)
            self.request.response.redirect(url)

    @button.buttonAndHandler(_("Cancel"))
    def handle_cancel_action(self, action):
        url = absoluteURL(self.context, self.request)
        self.request.response.redirect(url)


class FlourishSectionDeleteView(DialogForm, form.EditForm):
    """View used for confirming deletion of a section."""

    dialog_submit_actions = ('apply',)
    dialog_close_actions = ('cancel',)
    label = None

    @button.buttonAndHandler(_("Delete"), name='apply')
    def handleDelete(self, action):
        url = '%s/delete.html?delete.%s&CONFIRM' % (
            absoluteURL(self.context.__parent__, self.request),
            self.context.__name__.encode('utf-8'))
        self.request.response.redirect(url)
        # We never have errors, so just close the dialog.
        self.ajax_settings['dialog'] = 'close'

    @button.buttonAndHandler(_("Cancel"))
    def handle_cancel_action(self, action):
        pass

    def updateActions(self):
        super(FlourishSectionDeleteView, self).updateActions()
        self.actions['apply'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')


class FlourishSectionInstructorView(EditPersonRelationships):
    """View for adding instructors to a Section."""

    @property
    def title(self):
        return self.context.title

    current_title = _("Current instructors")
    available_title = _("Add instructors")

    def getSelectedItems(self):
        return filter(IPerson.providedBy, self.context.instructors)

    def getCollection(self):
        return self.context.instructors


class FlourishSectionLearnerView(EditPersonRelationships):
    """View for adding learners to a Section."""

    @property
    def title(self):
        return self.context.title

    current_title = _("Current students")
    available_title = _("Add students")

    def getSelectedItems(self):
        return filter(IPerson.providedBy, self.context.members)

    def getCollection(self):
        return self.context.members


class FlourishSectionLocationView(EditLocationRelationships):
    """View for adding locations to a Section."""

    @property
    def title(self):
        return self.context.title

    current_title = _("Current locations")
    available_title = _("Add locations")

    def getCollection(self):
        return self.context.resources

    def getSelectedItems(self):
        return [r for r in self.context.resources
                if ILocation(r, None) is not None]

    def getOmmitedItems(self):
        items = self.getSelectedItems()
        return [r for r in self.getAvailableItemsContainer().values()
                if ILocation(r, None) is None
                or r in items]


class FlourishSectionEquipmentView(EditEquipmentRelationships):
    """View for adding equipment to a Section."""

    @property
    def title(self):
        return self.context.title

    current_title = _("Current equipment")
    available_title = _("Add equipment")

    def getCollection(self):
        return self.context.resources

    def getSelectedItems(self):
        return [r for r in self.context.resources
                if IEquipment(r, None) is not None]

    def getOmmitedItems(self):
        items = self.getSelectedItems()
        return [r for r in self.getAvailableItemsContainer().values()
                if IEquipment(r, None) is None
                or r in items]


class FlourishSectionLinkageView(Page, SectionLinkageView):

    @property
    def columns(self):
        linked_sections = dict([(ITerm(section), section)
                                 for section in self.context.linked_sections])
        columns = []
        for term in sorted(self.year.values(), key=lambda t: t.first):
            section = linked_sections.get(term)
            current = sameProxiedObjects(section, self.context)
            info = {
                'term': term,
                'section': section,
                'current': current,
                'link_id': term.__name__.replace('.', '_'),
                'form_id': term.__name__.replace('.', '_') + '_container',
                }
            if section is not None:
                unlink_form_url = '%s/unlink_section.html' % (
                    absoluteURL(self.context, self.request))
                info['unlink_form_url'] = unlink_form_url
                info['unlink_dialog_title'] = self.unlink_dialog_title(term)
            else:
                extend_form_url = '%s/extend_term.html?term=%s' % (
                    absoluteURL(self.context, self.request),
                    term.__name__)
                link_existing_form_url = '%s/link_existing.html?term=%s' % (
                    absoluteURL(self.context, self.request),
                    term.__name__)
                link_existing_link_id = 'existing_' + info['link_id']
                link_existing_form_id = link_existing_link_id + '_container'
                link_existing_dialog_title = self.link_existing_dialog_title()
                info.update(
                    {'extend_form_url': extend_form_url,
                     'extend_dialog_title': self.extend_dialog_title(term),
                     'link_existing_form_url': link_existing_form_url,
                     'link_existing_link_id': link_existing_link_id,
                     'link_existing_form_id': link_existing_form_id,
                     'link_existing_dialog_title': link_existing_dialog_title,
                     })
            columns.append(info)
        return columns

    @property
    def done_link(self):
        return absoluteURL(self.context, self.request)

    def extend_dialog_title(self, term):
        title = _('Extend ${section} to ${term}',
                  mapping={'section': self.context.title,
                           'term': term.title})
        return translate(title, context=self.request)

    def unlink_dialog_title(self, term):
        title = _('Remove links of ${section}',
                  mapping={'section': self.context.title})
        return translate(title, context=self.request)

    def link_existing_dialog_title(self):
        title = _('Link ${section} to Section in Other Term',
                  mapping={'section': self.context.title})
        return translate(title, context=self.request)


class FlourishExtendTermView(Dialog, ExtendTermView):

    def update(self):
        Dialog.update(self)

        section = removeSecurityProxy(self.context)
        year = ISchoolYear(section)
        linked_names = [ITerm(s).__name__ for s in section.linked_sections]
        key = self.request['term']

        if key not in year or key in linked_names or 'CANCEL' in self.request:
            self.request.response.redirect(self.nextURL())

        elif 'EXTEND' in self.request:
            this_term = ITerm(self.context)
            extend_term = year[key]
            if extend_term.first < this_term.first:
                current = section.linked_sections[0]
                target_term = getPreviousTerm(ITerm(current))
            else:
                current = section.linked_sections[-1]
                target_term = getNextTerm(ITerm(current))
            while ITerm(current).first != extend_term.first:
                new_section = copySection(current, target_term)
                if extend_term.first < this_term.first:
                    new_section.next = current
                    target_term = getPreviousTerm(target_term)
                else:
                    new_section.previous = current
                    target_term = getNextTerm(target_term)
                current = new_section
            self.request.response.redirect(self.nextURL())


class FlourishUnlinkSectionView(Dialog, UnlinkSectionView):

    def update(self):
        Dialog.update(self)

        section = removeSecurityProxy(self.context)

        if not self.sections or 'CANCEL' in self.request:
            self.request.response.redirect(self.nextURL())

        elif 'UNLINK' in self.request:
            if section.previous:
                section.previous = None
            if section.next:
                section.next = None
            self.request.response.redirect(self.nextURL())


class FlourishLinkExistingView(Dialog, LinkExistingView):

    reload_parent = False

    def update(self):
        Dialog.update(self)

        section = removeSecurityProxy(self.context)
        term = self.term
        link_term = self.link_term

        if link_term is None or 'CANCEL' in self.request:
            self.reload_parent = True
            self.request.response.redirect(self.nextURL())

        elif 'LINK' in self.request:
            target_section = self.selected_section
            if target_section is None:
                return

            if link_term.first < term.first:
                current = section.linked_sections[0]
                target_term = getPreviousTerm(ITerm(current))
            else:
                current = section.linked_sections[-1]
                target_term = getNextTerm(ITerm(current))

            while target_term.first != link_term.first:
                new_section = copySection(current, target_term)
                if link_term.first < term.first:
                    new_section.next = current
                    target_term = getPreviousTerm(target_term)
                else:
                    new_section.previous = current
                    target_term = getNextTerm(target_term)
                current = new_section

            if link_term.first < term.first:
                target_section.next = current
            else:
                target_section.previous = current

            self.reload_parent = True
            self.request.response.redirect(self.nextURL())


class SectionsYearNavBreadcrumbs(SchoolyearNavBreadcrumbs):

    traversal_name = u'sections'
    title = _('Sections')


class SectionRosterPDFView(flourish.report.PlainPDFPage):

    name = _("Section Roster")

    content_template = flourish.templates.XMLFile('rml/section_roster.pt')

    @property
    def message_title(self):
        return _("section ${title} roster",
                 mapping={'title': self.context.title})

    def formatDate(self, date, format='mediumDate'):
        if date is None:
            return ''
        formatter = getMultiAdapter((date, self.request), name=format)
        return formatter()

    @property
    def scope(self):
        term = ITerm(self.context)
        first = self.formatDate(term.first)
        last = self.formatDate(term.last)
        return '%s | %s - %s' % (term.title, first, last)

    @property
    def subtitles_left(self):
        section = removeSecurityProxy(self.context)
        instructors = '; '.join([person.title
                                 for person in section.instructors])
        instructors_message = _('Instructors: ${instructors}',
                                mapping={'instructors': instructors})
        subtitles = [
            '%s (%s)' % (section.title, section.__name__),
            instructors_message,
            ]
        return subtitles

    @property
    def title(self):
        return ', '.join([course.title for course in self.context.courses])

    @property
    def base_filename(self):
        courses = [c.__name__ for c in self.context.courses]
        return 'section_roster_%s' % '_'.join(courses)

    @property
    def term(self):
        return ITerm(self.context)

    def rows(self):
        result = []
        collator = ICollator(self.request.locale)
        factory = getUtility(IPersonFactory)
        sorting_key = lambda x: factory.getSortingKey(x, collator)
        for student in sorted(self.context.members, key=sorting_key):
            demographics = IDemographics(student)
            result.append({
                    'full_name': self.full_name(student),
                    'ID': demographics.get('ID', '')
                    })
        return result

    def full_name(self, person):
        return person.title


class FlourishRequestSectionRosterView(RequestRemoteReportDialog):

    report_builder = SectionRosterPDFView
